#!/usr/bin/python
"""
Step file creator/editor.

:copyright: Red Hat Inc 2009
:author: mgoldish@redhat.com (Michael Goldish)
"""

import os
import glob
import shutil
import sys
import logging

try:
    from gi import pygtkcompat as pygtk
except ImportError:
    pygtk = None
if pygtk is not None:
    pygtk.enable()
    pygtk.enable_gtk(version='3.0')
import gtk

from virttest import ppm_utils


# General utilities

def corner_and_size_clipped(startpoint, endpoint, width, height):
    limits = width, height
    c0 = startpoint[:]
    c1 = endpoint[:]
    if c0[0] < 0:
        c0[0] = 0
    if c0[1] < 0:
        c0[1] = 0
    if c1[0] < 0:
        c1[0] = 0
    if c1[1] < 0:
        c1[1] = 0
    if c0[0] > limits[0] - 1:
        c0[0] = limits[0] - 1
    if c0[1] > limits[1] - 1:
        c0[1] = limits[1] - 1
    if c1[0] > limits[0] - 1:
        c1[0] = limits[0] - 1
    if c1[1] > limits[1] - 1:
        c1[1] = limits[1] - 1
    return ([min(c0[0], c1[0]),
             min(c0[1], c1[1])],
            [abs(c1[0] - c0[0]) + 1,
             abs(c1[1] - c0[1]) + 1])


def key_event_to_qemu_string(event):
    keymap = gtk.gdk.Keymap.get_default()
    keyvals = keymap.get_entries_for_keycode(event.hardware_keycode)
    keyval = keyvals[2][0]
    keyname = gtk.gdk.keyval_name(keyval)

    keymap = {"Return": "ret",
              "Tab": "tab",
              "space": "spc",
              "Left": "left",
              "Right": "right",
              "Up": "up",
              "Down": "down",
              "F1": "f1",
              "F2": "f2",
              "F3": "f3",
              "F4": "f4",
              "F5": "f5",
              "F6": "f6",
              "F7": "f7",
              "F8": "f8",
              "F9": "f9",
              "F10": "f10",
              "F11": "f11",
              "F12": "f12",
              "Escape": "esc",
              "minus": "minus",
              "equal": "equal",
              "BackSpace": "backspace",
              "comma": "comma",
              "period": "dot",
              "slash": "slash",
              "Insert": "insert",
              "Delete": "delete",
              "Home": "home",
              "End": "end",
              "Page_Up": "pgup",
              "Page_Down": "pgdn",
              "Menu": "menu",
              "semicolon": "0x27",
              "backslash": "0x2b",
              "apostrophe": "0x28",
              "grave": "0x29",
              "less": "0x2b",
              "bracketleft": "0x1a",
              "bracketright": "0x1b",
              "Super_L": "0xdc",
              "Super_R": "0xdb",
              }

    if ord('a') <= keyval <= ord('z') or ord('0') <= keyval <= ord('9'):
        sr = keyname
    elif keyname in list(keymap.keys()):
        sr = keymap[keyname]
    else:
        return ""

    if event.state & gtk.gdk.CONTROL_MASK:
        sr = "ctrl-" + sr
    if event.state & gtk.gdk.MOD1_MASK:
        sr = "alt-" + sr
    if event.state & gtk.gdk.SHIFT_MASK:
        sr = "shift-" + sr

    return sr


class StepMakerWindow(object):

    def __init__(self):
        # Window
        self.window = gtk.Window(gtk.WINDOW_TOPLEVEL)
        self.window.set_title("Step Maker Window")
        self.window.connect("delete-event", self.delete_event)
        self.window.connect("destroy", self.destroy)
        self.window.set_default_size(600, 800)

        # Main box (inside a frame which is inside a VBox)
        self.menu_vbox = gtk.VBox()
        self.window.add(self.menu_vbox)
        self.menu_vbox.show()

        frame = gtk.Frame()
        frame.set_border_width(10)
        frame.set_shadow_type(gtk.SHADOW_NONE)
        self.menu_vbox.pack_end(frame)
        frame.show()

        self.main_vbox = gtk.VBox(spacing=10)
        frame.add(self.main_vbox)
        self.main_vbox.show()

        # EventBox
        self.scrolledwindow = gtk.ScrolledWindow()
        self.scrolledwindow.set_policy(gtk.POLICY_AUTOMATIC,
                                       gtk.POLICY_AUTOMATIC)
        self.scrolledwindow.set_shadow_type(gtk.SHADOW_NONE)
        self.main_vbox.pack_start(self.scrolledwindow)
        self.scrolledwindow.show()

        table = gtk.Table(1, 1)
        self.scrolledwindow.add_with_viewport(table)
        table.show()
        table.realize()

        self.event_box = gtk.EventBox()
        table.attach(self.event_box, 0, 1, 0, 1, gtk.EXPAND, gtk.EXPAND)
        self.event_box.show()
        self.event_box.realize()

        # Image
        self.image = gtk.Image()
        self.event_box.add(self.image)
        self.image.show()

        # Data VBox
        self.data_vbox = gtk.VBox(spacing=10)
        self.main_vbox.pack_start(self.data_vbox, expand=False)
        self.data_vbox.show()

        # User VBox
        self.user_vbox = gtk.VBox(spacing=10)
        self.main_vbox.pack_start(self.user_vbox, expand=False)
        self.user_vbox.show()

        # Screendump ID HBox
        box = gtk.HBox(spacing=10)
        self.data_vbox.pack_start(box)
        box.show()

        label = gtk.Label("Screendump ID:")
        box.pack_start(label, False)
        label.show()

        self.entry_screendump = gtk.Entry()
        self.entry_screendump.set_editable(False)
        box.pack_start(self.entry_screendump)
        self.entry_screendump.show()

        label = gtk.Label("Time:")
        box.pack_start(label, False)
        label.show()

        self.entry_time = gtk.Entry()
        self.entry_time.set_editable(False)
        self.entry_time.set_width_chars(10)
        box.pack_start(self.entry_time, False)
        self.entry_time.show()

        # Comment HBox
        box = gtk.HBox(spacing=10)
        self.data_vbox.pack_start(box)
        box.show()

        label = gtk.Label("Comment:")
        box.pack_start(label, False)
        label.show()

        self.entry_comment = gtk.Entry()
        box.pack_start(self.entry_comment)
        self.entry_comment.show()

        # Sleep HBox
        box = gtk.HBox(spacing=10)
        self.data_vbox.pack_start(box)
        box.show()

        self.check_sleep = gtk.CheckButton("Sleep:")
        self.check_sleep.connect("toggled", self.event_check_sleep_toggled)
        box.pack_start(self.check_sleep, False)
        self.check_sleep.show()

        self.spin_sleep = gtk.SpinButton()
        adjustment = gtk.Adjustment(0, 0, 50000, 1, 10, 0)
        self.spin_sleep.set_adjustment(adjustment)
        self.spin_sleep.climb_rate = 0.0
        box.pack_start(self.spin_sleep, False)
        self.spin_sleep.show()

        # Barrier HBox
        box = gtk.HBox(spacing=10)
        self.data_vbox.pack_start(box)
        box.show()

        self.check_barrier = gtk.CheckButton("Barrier:")
        self.check_barrier.connect("toggled", self.event_check_barrier_toggled)
        box.pack_start(self.check_barrier, False)
        self.check_barrier.show()

        vbox = gtk.VBox()
        box.pack_start(vbox)
        vbox.show()

        self.label_barrier_region = gtk.Label("Region:")
        self.label_barrier_region.set_alignment(0, 0.5)
        vbox.pack_start(self.label_barrier_region)
        self.label_barrier_region.show()

        self.label_barrier_md5sum = gtk.Label("MD5:")
        self.label_barrier_md5sum.set_alignment(0, 0.5)
        vbox.pack_start(self.label_barrier_md5sum)
        self.label_barrier_md5sum.show()

        self.label_barrier_timeout = gtk.Label("Timeout:")
        box.pack_start(self.label_barrier_timeout, False)
        self.label_barrier_timeout.show()

        self.spin_barrier_timeout = gtk.SpinButton()
        adjustment = gtk.Adjustment(0, 0, 50000, 1, 10, 0)
        self.spin_barrier_timeout.set_adjustment(adjustment)
        self.spin_barrier_timeout.climb_rate = 0.0
        box.pack_start(self.spin_barrier_timeout, False)
        self.spin_barrier_timeout.show()

        self.check_barrier_optional = gtk.CheckButton("Optional")
        box.pack_start(self.check_barrier_optional, False)
        self.check_barrier_optional.show()

        # Keystrokes HBox
        box = gtk.HBox(spacing=10)
        self.data_vbox.pack_start(box)
        box.show()

        label = gtk.Label("Keystrokes:")
        box.pack_start(label, False)
        label.show()

        frame = gtk.Frame()
        frame.set_shadow_type(gtk.SHADOW_IN)
        box.pack_start(frame)
        frame.show()

        self.entry_keys = gtk.TextView()
        self.text_buffer = self.entry_keys.get_buffer()
        self.entry_keys.set_wrap_mode(gtk.WRAP_WORD)
        self.entry_keys.connect("key-press-event", self.event_key_press)
        frame.add(self.entry_keys)
        self.entry_keys.show()

        self.check_manual = gtk.CheckButton("Manual")
        self.check_manual.connect("toggled", self.event_manual_toggled)
        box.pack_start(self.check_manual, False)
        self.check_manual.show()

        button = gtk.Button("Clear")
        button.connect("clicked", self.event_clear_clicked)
        box.pack_start(button, False)
        button.show()

        # Mouse click HBox
        box = gtk.HBox(spacing=10)
        self.data_vbox.pack_start(box)
        box.show()

        label = gtk.Label("Mouse action:")
        box.pack_start(label, False)
        label.show()

        self.button_capture = gtk.Button("Capture")
        box.pack_start(self.button_capture, False)
        self.button_capture.show()

        self.check_mousemove = gtk.CheckButton("Move: ...")
        box.pack_start(self.check_mousemove, False)
        self.check_mousemove.show()

        self.check_mouseclick = gtk.CheckButton("Click: ...")
        box.pack_start(self.check_mouseclick, False)
        self.check_mouseclick.show()

        self.spin_sensitivity = gtk.SpinButton()
        adjustment = gtk.Adjustment(1, 1, 100, 1, 10, 0)
        self.spin_sensitivity.set_adjustment(adjustment)
        self.spin_sensitivity.climb_rate = 0.0
        box.pack_end(self.spin_sensitivity, False)
        self.spin_sensitivity.show()

        label = gtk.Label("Sensitivity:")
        box.pack_end(label, False)
        label.show()

        self.spin_latency = gtk.SpinButton()
        adjustment = gtk.Adjustment(10, 1, 500, 1, 10, 0)
        self.spin_latency.set_adjustment(adjustment)
        self.spin_latency.climb_rate = 0.0
        box.pack_end(self.spin_latency, False)
        self.spin_latency.show()

        label = gtk.Label("Latency:")
        box.pack_end(label, False)
        label.show()

        self.handler_event_box_press = None
        self.handler_event_box_release = None
        self.handler_event_box_scroll = None
        self.handler_event_box_motion = None
        self.handler_event_box_draw = None

        self.window.realize()
        # maximize to handle bug where the barrier region is not drawn
        self.window.maximize()
        self.window.show()

        self.clear_state()

    # Utilities

    def message(self, text, title):
        dlg = gtk.MessageDialog(self.window,
                                gtk.DIALOG_MODAL | gtk.DIALOG_DESTROY_WITH_PARENT,
                                gtk.MESSAGE_INFO,
                                gtk.BUTTONS_CLOSE,
                                title)
        dlg.set_title(title)
        dlg.format_secondary_text(text)
        dlg.run()
        dlg.destroy()

    def question_yes_no(self, text, title):
        dlg = gtk.MessageDialog(self.window,
                                gtk.DIALOG_MODAL | gtk.DIALOG_DESTROY_WITH_PARENT,
                                gtk.MESSAGE_QUESTION,
                                gtk.BUTTONS_YES_NO,
                                title)
        dlg.set_title(title)
        dlg.format_secondary_text(text)
        response = dlg.run()
        dlg.destroy()
        if response == gtk.RESPONSE_YES:
            return True
        return False

    def inputdialog(self, text, title, default_response=""):
        # Define a little helper function
        def inputdialog_entry_activated(entry):
            dlg.response(gtk.RESPONSE_OK)

        # Create the dialog
        dlg = gtk.MessageDialog(self.window,
                                gtk.DIALOG_MODAL | gtk.DIALOG_DESTROY_WITH_PARENT,
                                gtk.MESSAGE_QUESTION,
                                gtk.BUTTONS_OK_CANCEL,
                                title)
        dlg.set_title(title)
        dlg.format_secondary_text(text)

        # Create an entry widget
        entry = gtk.Entry()
        entry.set_text(default_response)
        entry.connect("activate", inputdialog_entry_activated)
        dlg.vbox.pack_start(entry)
        entry.show()

        # Run the dialog
        response = dlg.run()
        dlg.destroy()
        if response == gtk.RESPONSE_OK:
            return entry.get_text()
        return None

    def filedialog(self, title=None, default_filename=None):
        chooser = gtk.FileChooserDialog(title=title, parent=self.window,
                                        action=gtk.FILE_CHOOSER_ACTION_OPEN,
                                        buttons=(
                                            gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL, gtk.STOCK_OPEN,
                                            gtk.RESPONSE_OK))
        chooser.resize(700, 500)
        if default_filename:
            chooser.set_filename(os.path.abspath(default_filename))
        filename = None
        response = chooser.run()
        if response == gtk.RESPONSE_OK:
            filename = chooser.get_filename()
        chooser.destroy()
        return filename

    def redirect_event_box_input(self, press=None, release=None, scroll=None,
                                 motion=None, draw=None):
        if self.handler_event_box_press is not None:
            self.event_box.disconnect(self.handler_event_box_press)
        if self.handler_event_box_release is not None:
            self.event_box.disconnect(self.handler_event_box_release)
        if self.handler_event_box_scroll is not None:
            self.event_box.disconnect(self.handler_event_box_scroll)
        if self.handler_event_box_motion is not None:
            self.event_box.disconnect(self.handler_event_box_motion)
        if self.handler_event_box_draw is not None:
            self.event_box.disconnect(self.handler_event_box_draw)
        self.handler_event_box_press = None
        self.handler_event_box_release = None
        self.handler_event_box_scroll = None
        self.handler_event_box_motion = None
        self.handler_event_box_draw = None
        if press is not None:
            self.handler_event_box_press = (
                self.event_box.connect("button-press-event", press))
        if release is not None:
            self.handler_event_box_release = (
                self.event_box.connect("button-release-event", release))
        if scroll is not None:
            self.handler_event_box_scroll = (
                self.event_box.connect("scroll-event", scroll))
        if motion is not None:
            self.handler_event_box_motion = (
                self.event_box.connect("motion-notify-event", motion))
        if draw is not None:
            self.handler_event_box_draw = (
                self.event_box.connect_after("draw", draw))

    def get_keys(self):
        return self.text_buffer.get_text(
            self.text_buffer.get_start_iter(),
            self.text_buffer.get_end_iter(), True)

    def add_key(self, key):
        text = self.get_keys()
        if len(text) > 0 and text[-1] != ' ':
            text += " "
        text += key
        self.text_buffer.set_text(text)

    def clear_keys(self):
        self.text_buffer.set_text("")

    def update_barrier_info(self):
        if self.barrier_selected:
            self.label_barrier_region.set_text("Selected region: Corner: " +
                                               str(tuple(self.barrier_corner)) +
                                               " Size: " +
                                               str(tuple(self.barrier_size)))
        else:
            self.label_barrier_region.set_text("No region selected.")
        self.label_barrier_md5sum.set_text("MD5: " + self.barrier_md5sum)

    def update_mouse_click_info(self):
        if self.mouse_click_captured:
            self.check_mousemove.set_label("Move: " +
                                           str(tuple(self.mouse_click_coords)))
            self.check_mouseclick.set_label("Click: button %d" %
                                            self.mouse_click_button)
        else:
            self.check_mousemove.set_label("Move: ...")
            self.check_mouseclick.set_label("Click: ...")

    def clear_state(self, clear_screendump=True):
        # Recording time
        self.entry_time.set_text("unknown")
        if clear_screendump:
            # Screendump
            self.clear_image()
        # Screendump ID
        self.entry_screendump.set_text("")
        # Comment
        self.entry_comment.set_text("")
        # Sleep
        self.check_sleep.set_active(True)
        self.check_sleep.set_active(False)
        self.spin_sleep.set_value(10)
        # Barrier
        self.clear_barrier_state()
        # Keystrokes
        self.check_manual.set_active(False)
        self.clear_keys()
        # Mouse actions
        self.check_mousemove.set_sensitive(False)
        self.check_mouseclick.set_sensitive(False)
        self.check_mousemove.set_active(False)
        self.check_mouseclick.set_active(False)
        self.mouse_click_captured = False
        self.mouse_click_coords = [0, 0]
        self.mouse_click_button = 0
        self.update_mouse_click_info()

    def clear_barrier_state(self):
        self.check_barrier.set_active(True)
        self.check_barrier.set_active(False)
        self.check_barrier_optional.set_active(False)
        self.spin_barrier_timeout.set_value(10)
        self.barrier_selection_started = False
        self.barrier_selected = False
        self.barrier_corner0 = [0, 0]
        self.barrier_corner1 = [0, 0]
        self.barrier_corner = [0, 0]
        self.barrier_size = [0, 0]
        self.barrier_md5sum = ""
        self.update_barrier_info()

    def set_image(self, w, h, data):
        (self.image_width, self.image_height, self.image_data) = (w, h, data)
        self.image.set_from_pixbuf(gtk.gdk.pixbuf_new_from_data(
            data, gtk.gdk.COLORSPACE_RGB, False, 8,
            w, h, w * 3))
        hscrollbar = self.scrolledwindow.get_hscrollbar()
        hscrollbar.set_range(0, w)
        vscrollbar = self.scrolledwindow.get_vscrollbar()
        vscrollbar.set_range(0, h)

    def set_image_from_file(self, filename):
        if not ppm_utils.image_verify_ppm_file(filename):
            logging.warning("set_image_from_file: Warning: received invalid"
                            "screendump file")
            return self.clear_image()
        (w, h, data) = ppm_utils.image_read_from_ppm_file(filename)
        self.set_image(w, h, data)

    def clear_image(self):
        self.image.clear()
        self.image_width = 0
        self.image_height = 0
        self.image_data = b""

    def update_screendump_id(self, data_dir):
        if not self.image_data:
            return
        # Find a proper ID for the screendump
        scrdump_md5sum = ppm_utils.image_md5sum(self.image_width,
                                                self.image_height,
                                                self.image_data)
        scrdump_id = ppm_utils.find_id_for_screendump(scrdump_md5sum, data_dir)
        if not scrdump_id:
            # Not found; generate one
            scrdump_id = ppm_utils.generate_id_for_screendump(scrdump_md5sum,
                                                              data_dir)
        self.entry_screendump.set_text(scrdump_id)

    def get_step_lines(self, data_dir=None):
        if self.check_barrier.get_active() and not self.barrier_selected:
            self.message("No barrier region selected.", "Error")
            return

        sr = "step"

        # Add step recording time
        if self.entry_time.get_text():
            sr += " " + self.entry_time.get_text()

        sr += "\n"

        # Add screendump line
        if self.image_data:
            sr += "screendump %s\n" % self.entry_screendump.get_text()

        # Add comment
        if self.entry_comment.get_text():
            sr += "# %s\n" % self.entry_comment.get_text()

        # Add sleep line
        if self.check_sleep.get_active():
            sr += "sleep %d\n" % self.spin_sleep.get_value()

        # Add barrier_2 line
        if self.check_barrier.get_active():
            sr += "barrier_2 %d %d %d %d %s %d" % (
                self.barrier_size[0], self.barrier_size[1],
                self.barrier_corner[0], self.barrier_corner[1],
                self.barrier_md5sum, self.spin_barrier_timeout.get_value())
            if self.check_barrier_optional.get_active():
                sr += " optional"
            sr += "\n"

        # Add "Sending keys" comment
        keys_to_send = self.get_keys().split()
        if keys_to_send:
            sr += "# Sending keys: %s\n" % self.get_keys()

        # Add key and var lines
        for key in keys_to_send:
            if key.startswith("$"):
                varname = key[1:]
                sr += "var %s\n" % varname
            else:
                sr += "key %s\n" % key

        # Add mousemove line
        if self.check_mousemove.get_active():
            sr += "mousemove %d %d\n" % (self.mouse_click_coords[0],
                                         self.mouse_click_coords[1])

        # Add mouseclick line
        if self.check_mouseclick.get_active():
            mapping = {1: 1,
                       2: 2,
                       3: 4}
            sr += "mouseclick %d\n" % mapping[self.mouse_click_button]

        # Write screendump and cropped screendump image files
        if data_dir and self.image_data:
            # Create the data dir if it doesn't exist
            if not os.path.exists(data_dir):
                os.makedirs(data_dir)
            # Get the full screendump filename
            scrdump_filename = os.path.join(data_dir,
                                            self.entry_screendump.get_text())
            # Write screendump file if it doesn't exist
            if not os.path.exists(scrdump_filename):
                try:
                    ppm_utils.image_write_to_ppm_file(scrdump_filename,
                                                      self.image_width,
                                                      self.image_height,
                                                      self.image_data)
                except IOError:
                    self.message("Could not write screendump file.", "Error")

        return sr

    def set_state_from_step_lines(self, sr, data_dir, warn=True):
        self.clear_state()

        for line in sr.splitlines():
            words = line.split()
            if not words:
                continue

            if (line.startswith("#") and not
                self.entry_comment.get_text() and not
                line.startswith("# Sending keys:") and not
                    line.startswith("# ----")):

                self.entry_comment.set_text(line.strip("#").strip())

            elif words[0] == "step":
                if len(words) >= 2:
                    self.entry_time.set_text(words[1])

            elif words[0] == "screendump":
                self.entry_screendump.set_text(words[1])
                self.set_image_from_file(os.path.join(data_dir, words[1]))

            elif words[0] == "sleep":
                self.spin_sleep.set_value(int(words[1]))
                self.check_sleep.set_active(True)

            elif words[0] == "key":
                self.add_key(words[1])

            elif words[0] == "var":
                self.add_key("$%s" % words[1])

            elif words[0] == "mousemove":
                self.mouse_click_captured = True
                self.mouse_click_coords = [int(words[1]), int(words[2])]
                self.update_mouse_click_info()

            elif words[0] == "mouseclick":
                self.mouse_click_captured = True
                self.mouse_click_button = int(words[1])
                self.update_mouse_click_info()

            elif words[0] == "barrier_2":
                # Get region corner and size from step lines
                self.barrier_corner = [int(words[3]), int(words[4])]
                self.barrier_size = [int(words[1]), int(words[2])]
                # Get corner0 and corner1 from step lines
                self.barrier_corner0 = self.barrier_corner
                self.barrier_corner1 = [self.barrier_corner[0] +
                                        self.barrier_size[0] - 1,
                                        self.barrier_corner[1] +
                                        self.barrier_size[1] - 1]
                # Get the md5sum
                self.barrier_md5sum = words[5]
                # Pretend the user selected the region with the mouse
                self.barrier_selection_started = True
                self.barrier_selected = True
                # Update label widgets according to region information
                self.update_barrier_info()
                # Check the barrier checkbutton
                self.check_barrier.set_active(True)
                # Set timeout value
                self.spin_barrier_timeout.set_value(int(words[6]))
                # Set 'optional' checkbutton state
                self.check_barrier_optional.set_active(words[-1] == "optional")
                # Update the image widget
                self.event_box.queue_draw()

                if warn:
                    # See if the computed md5sum matches the one recorded in
                    # the file
                    computed_md5sum = ppm_utils.get_region_md5sum(
                        self.image_width, self.image_height,
                        self.image_data, self.barrier_corner[0],
                        self.barrier_corner[1], self.barrier_size[0],
                        self.barrier_size[1])
                    if computed_md5sum != self.barrier_md5sum:
                        self.message("Computed MD5 sum (%s) differs from MD5"
                                     " sum recorded in steps file (%s)" %
                                     (computed_md5sum, self.barrier_md5sum),
                                     "Warning")

    # Events

    def delete_event(self, widget, event):
        pass

    def destroy(self, widget):
        gtk.main_quit()

    def event_check_barrier_toggled(self, widget):
        if self.check_barrier.get_active():
            self.redirect_event_box_input(
                self.event_button_press,
                self.event_button_release,
                None,
                None,
                self.event_draw)
            self.event_box.queue_draw()
            cursor = gtk.gdk.Cursor(gtk.gdk.CursorType.CROSSHAIR)
            self.event_box.window.set_cursor(cursor)
            self.label_barrier_region.set_sensitive(True)
            self.label_barrier_md5sum.set_sensitive(True)
            self.label_barrier_timeout.set_sensitive(True)
            self.spin_barrier_timeout.set_sensitive(True)
            self.check_barrier_optional.set_sensitive(True)
        else:
            self.redirect_event_box_input()
            self.event_box.queue_draw()
            self.event_box.window.set_cursor(None)
            self.label_barrier_region.set_sensitive(False)
            self.label_barrier_md5sum.set_sensitive(False)
            self.label_barrier_timeout.set_sensitive(False)
            self.spin_barrier_timeout.set_sensitive(False)
            self.check_barrier_optional.set_sensitive(False)

    def event_check_sleep_toggled(self, widget):
        if self.check_sleep.get_active():
            self.spin_sleep.set_sensitive(True)
        else:
            self.spin_sleep.set_sensitive(False)

    def event_manual_toggled(self, widget):
        self.entry_keys.grab_focus()

    def event_clear_clicked(self, widget):
        self.clear_keys()
        self.entry_keys.grab_focus()

    def event_draw(self, widget, event):
        if not self.barrier_selection_started:
            return
        max_width = self.event_box.get_preferred_width()[1]
        max_height = self.event_box.get_preferred_height()[1]
        (corner, size) = corner_and_size_clipped(self.barrier_corner0,
                                                 self.barrier_corner1,
                                                 max_width, max_height)
        cr = self.event_box.window.cairo_create()
        cr.set_source_rgb(1.0, 0.0, 0.0)
        cr.set_line_width(1.0)
        cr.set_dash((4.0, 4.0), 0)
        cr.rectangle(corner[0], corner[1], size[0] - 1, size[1] - 1)
        cr.stroke()

    def event_drag_motion(self, widget, event):
        old_corner1 = self.barrier_corner1
        self.barrier_corner1 = [int(event.x), int(event.y)]
        max_width = self.event_box.get_preferred_width()[1]
        max_height = self.event_box.get_preferred_height()[1]
        (corner, size) = corner_and_size_clipped(self.barrier_corner0,
                                                 self.barrier_corner1,
                                                 max_width, max_height)
        (old_corner, old_size) = corner_and_size_clipped(self.barrier_corner0,
                                                         old_corner1,
                                                         max_width, max_height)
        corner0 = [
            min(corner[0], old_corner[0]), min(corner[1], old_corner[1])]
        corner1 = [max(corner[0] + size[0], old_corner[0] + old_size[0]),
                   max(corner[1] + size[1], old_corner[1] + old_size[1])]
        size = [corner1[0] - corner0[0] + 1,
                corner1[1] - corner0[1] + 1]
        self.event_box.queue_draw_area(
            corner0[0], corner0[1], size[0], size[1])

    def event_button_press(self, widget, event):
        max_width = self.event_box.get_preferred_width()[1]
        max_height = self.event_box.get_preferred_height()[1]
        (corner, size) = corner_and_size_clipped(self.barrier_corner0,
                                                 self.barrier_corner1,
                                                 max_width, max_height)
        self.event_box.queue_draw_area(corner[0], corner[1], size[0], size[1])
        self.barrier_corner0 = [int(event.x), int(event.y)]
        self.barrier_corner1 = [int(event.x), int(event.y)]
        self.redirect_event_box_input(
            self.event_button_press,
            self.event_button_release,
            None,
            self.event_drag_motion,
            self.event_draw)
        self.barrier_selection_started = True

    def event_button_release(self, widget, event):
        max_width = self.event_box.get_preferred_width()[1]
        max_height = self.event_box.get_preferred_height()[1]
        self.redirect_event_box_input(
            self.event_button_press,
            self.event_button_release,
            None,
            None,
            self.event_draw)
        (self.barrier_corner, self.barrier_size) = \
            corner_and_size_clipped(self.barrier_corner0, self.barrier_corner1,
                                    max_width, max_height)
        self.barrier_md5sum = ppm_utils.get_region_md5sum(
            self.image_width, self.image_height, self.image_data,
            self.barrier_corner[0], self.barrier_corner[1],
            self.barrier_size[0], self.barrier_size[1])
        self.barrier_selected = True
        self.update_barrier_info()

    def event_key_press(self, widget, event):
        if self.check_manual.get_active():
            return False
        sr = key_event_to_qemu_string(event)
        self.add_key(sr)
        return True


class StepEditor(StepMakerWindow):
    ui = '''<ui>
    <menubar name="MenuBar">
        <menu action="File">
            <menuitem action="Open"/>
            <separator/>
            <menuitem action="Quit"/>
        </menu>
        <menu action="Edit">
            <menuitem action="CopyStep"/>
            <menuitem action="DeleteStep"/>
        </menu>
        <menu action="Insert">
            <menuitem action="InsertNewBefore"/>
            <menuitem action="InsertNewAfter"/>
            <separator/>
            <menuitem action="InsertStepsBefore"/>
            <menuitem action="InsertStepsAfter"/>
        </menu>
        <menu action="Tools">
            <menuitem action="CleanUp"/>
        </menu>
    </menubar>
</ui>'''

    # Constructor

    def __init__(self, filename=None):
        StepMakerWindow.__init__(self)

        self.steps_filename = None
        self.steps = []

        # Create a UIManager instance
        uimanager = gtk.UIManager()

        # Add the accelerator group to the toplevel window
        accelgroup = uimanager.get_accel_group()
        self.window.add_accel_group(accelgroup)

        # Create an ActionGroup
        actiongroup = gtk.ActionGroup('StepEditor')

        # Create actions
        actiongroup.add_actions([
            ('Quit', gtk.STOCK_QUIT, '_Quit', None, 'Quit the Program',
             self.quit),
            ('Open', gtk.STOCK_OPEN, '_Open', None, 'Open steps file',
             self.open_steps_file),
            ('CopyStep', gtk.STOCK_COPY, '_Copy current step...', "",
             'Copy current step to user specified position', self.copy_step),
            ('DeleteStep', gtk.STOCK_DELETE, '_Delete current step', "",
             'Delete current step', self.event_remove_clicked),
            ('InsertNewBefore', gtk.STOCK_ADD, '_New step before current', "",
             'Insert new step before current step', self.insert_before),
            ('InsertNewAfter', gtk.STOCK_ADD, 'N_ew step after current', "",
             'Insert new step after current step', self.insert_after),
            ('InsertStepsBefore', gtk.STOCK_ADD, '_Steps before current...',
             "", 'Insert steps (from file) before current step',
             self.insert_steps_before),
            ('InsertStepsAfter', gtk.STOCK_ADD, 'Steps _after current...', "",
             'Insert steps (from file) after current step',
             self.insert_steps_after),
            ('CleanUp', gtk.STOCK_DELETE, '_Clean up data directory', "",
             'Move unused PPM files to a backup directory', self.cleanup),
            ('File', None, '_File'),
            ('Edit', None, '_Edit'),
            ('Insert', None, '_Insert'),
            ('Tools', None, '_Tools')
        ])

        def create_shortcut(name, callback, keyname):
            # Create an action
            action = gtk.Action(name, None, None, None)
            # Connect a callback to the action
            action.connect("activate", callback)
            actiongroup.add_action_with_accel(action, keyname)
            # Have the action use accelgroup
            action.set_accel_group(accelgroup)
            # Connect the accelerator to the action
            action.connect_accelerator()

        create_shortcut("Next", self.event_next_clicked, "Page_Down")
        create_shortcut("Previous", self.event_prev_clicked, "Page_Up")

        # Add the actiongroup to the uimanager
        uimanager.insert_action_group(actiongroup, 0)

        # Add a UI description
        uimanager.add_ui_from_string(self.ui)

        # Create a MenuBar
        menubar = uimanager.get_widget('/MenuBar')
        self.menu_vbox.pack_start(menubar, False)

        # Remember the Edit menu bar for future reference
        self.menu_edit = uimanager.get_widget('/MenuBar/Edit')
        self.menu_edit.set_sensitive(False)

        # Remember the Insert menu bar for future reference
        self.menu_insert = uimanager.get_widget('/MenuBar/Insert')
        self.menu_insert.set_sensitive(False)

        # Remember the Tools menu bar for future reference
        self.menu_tools = uimanager.get_widget('/MenuBar/Tools')
        self.menu_tools.set_sensitive(False)

        # Next/Previous HBox
        hbox = gtk.HBox(spacing=10)
        self.user_vbox.pack_start(hbox)
        hbox.show()

        self.button_first = gtk.Button(stock=gtk.STOCK_GOTO_FIRST)
        self.button_first.connect("clicked", self.event_first_clicked)
        hbox.pack_start(self.button_first)
        self.button_first.show()

        #self.button_prev = gtk.Button("<< Previous")
        self.button_prev = gtk.Button(stock=gtk.STOCK_GO_BACK)
        self.button_prev.connect("clicked", self.event_prev_clicked)
        hbox.pack_start(self.button_prev)
        self.button_prev.show()

        self.label_step = gtk.Label("Step:")
        hbox.pack_start(self.label_step, False)
        self.label_step.show()

        self.entry_step_num = gtk.Entry()
        self.entry_step_num.connect(
            "activate", self.event_entry_step_activated)
        self.entry_step_num.set_width_chars(3)
        hbox.pack_start(self.entry_step_num, False)
        self.entry_step_num.show()

        #self.button_next = gtk.Button("Next >>")
        self.button_next = gtk.Button(stock=gtk.STOCK_GO_FORWARD)
        self.button_next.connect("clicked", self.event_next_clicked)
        hbox.pack_start(self.button_next)
        self.button_next.show()

        self.button_last = gtk.Button(stock=gtk.STOCK_GOTO_LAST)
        self.button_last.connect("clicked", self.event_last_clicked)
        hbox.pack_start(self.button_last)
        self.button_last.show()

        # Save HBox
        hbox = gtk.HBox(spacing=10)
        self.user_vbox.pack_start(hbox)
        hbox.show()

        self.button_save = gtk.Button("_Save current step")
        self.button_save.connect("clicked", self.event_save_clicked)
        hbox.pack_start(self.button_save)
        self.button_save.show()

        self.button_remove = gtk.Button("_Delete current step")
        self.button_remove.connect("clicked", self.event_remove_clicked)
        hbox.pack_start(self.button_remove)
        self.button_remove.show()

        self.button_replace = gtk.Button("_Replace screendump")
        self.button_replace.connect("clicked", self.event_replace_clicked)
        hbox.pack_start(self.button_replace)
        self.button_replace.show()

        # Disable unused widgets
        self.button_capture.set_sensitive(False)
        self.spin_latency.set_sensitive(False)
        self.spin_sensitivity.set_sensitive(False)

        # Disable main vbox because no steps file is loaded
        self.main_vbox.set_sensitive(False)

        # Set title
        self.window.set_title("Step Editor")

    # Events

    def delete_event(self, widget, event):
        # Make sure the step is saved (if the user wants it to be)
        self.verify_save()

    def event_first_clicked(self, widget):
        if not self.steps:
            return
        # Make sure the step is saved (if the user wants it to be)
        self.verify_save()
        # Go to first step
        self.set_step(0)

    def event_last_clicked(self, widget):
        if not self.steps:
            return
        # Make sure the step is saved (if the user wants it to be)
        self.verify_save()
        # Go to last step
        self.set_step(len(self.steps) - 1)

    def event_prev_clicked(self, widget):
        if not self.steps:
            return
        # Make sure the step is saved (if the user wants it to be)
        self.verify_save()
        # Go to previous step
        index = self.current_step_index - 1
        if self.steps:
            index = index % len(self.steps)
        self.set_step(index)

    def event_next_clicked(self, widget):
        if not self.steps:
            return
        # Make sure the step is saved (if the user wants it to be)
        self.verify_save()
        # Go to next step
        index = self.current_step_index + 1
        if self.steps:
            index = index % len(self.steps)
        self.set_step(index)

    def event_entry_step_activated(self, widget):
        if not self.steps:
            return
        step_index = self.entry_step_num.get_text()
        if not step_index.isdigit():
            return
        step_index = int(step_index) - 1
        if step_index == self.current_step_index:
            return
        self.verify_save()
        self.set_step(step_index)

    def event_save_clicked(self, widget):
        if not self.steps:
            return
        self.save_step()

    def event_remove_clicked(self, widget):
        if not self.steps:
            return
        if not self.question_yes_no("This will modify the steps file."
                                    " Are you sure?", "Remove step?"):
            return
        # Remove step
        del self.steps[self.current_step_index]
        # Write changes to file
        self.write_steps_file(self.steps_filename)
        # Move to previous step
        self.set_step(self.current_step_index)

    def event_replace_clicked(self, widget):
        if not self.steps:
            return
        # Let the user choose a screendump file
        current_filename = os.path.join(self.steps_data_dir,
                                        self.entry_screendump.get_text())
        filename = self.filedialog("Choose PPM image file",
                                   default_filename=current_filename)
        if not filename:
            return
        if not ppm_utils.image_verify_ppm_file(filename):
            self.message("Not a valid PPM image file.", "Error")
            return
        self.clear_image()
        self.clear_barrier_state()
        self.set_image_from_file(filename)
        self.update_screendump_id(self.steps_data_dir)

    # Menu actions

    def open_steps_file(self, action):
        # Make sure the step is saved (if the user wants it to be)
        self.verify_save()
        # Let the user choose a steps file
        current_filename = self.steps_filename
        filename = self.filedialog("Open steps file",
                                   default_filename=current_filename)
        if not filename:
            return
        self.set_steps_file(filename)

    def quit(self, action):
        # Make sure the step is saved (if the user wants it to be)
        self.verify_save()
        # Quit
        gtk.main_quit()

    def copy_step(self, action):
        if not self.steps:
            return
        self.verify_save()
        self.set_step(self.current_step_index)
        # Get the desired position
        step_index = self.inputdialog("Copy step to position:",
                                      "Copy step",
                                      str(self.current_step_index + 2))
        if not step_index:
            return
        step_index = int(step_index) - 1
        # Get the lines of the current step
        step = self.steps[self.current_step_index]
        # Insert new step at position step_index
        self.steps.insert(step_index, step)
        # Go to new step
        self.set_step(step_index)
        # Write changes to disk
        self.write_steps_file(self.steps_filename)

    def insert_before(self, action):
        if not self.steps_filename:
            return
        if not self.question_yes_no("This will modify the steps file."
                                    " Are you sure?", "Insert new step?"):
            return
        self.verify_save()
        step_index = self.current_step_index
        # Get the lines of a blank step
        self.clear_state()
        step = self.get_step_lines()
        # Insert new step at position step_index
        self.steps.insert(step_index, step)
        # Go to new step
        self.set_step(step_index)
        # Write changes to disk
        self.write_steps_file(self.steps_filename)

    def insert_after(self, action):
        if not self.steps_filename:
            return
        if not self.question_yes_no("This will modify the steps file."
                                    " Are you sure?", "Insert new step?"):
            return
        self.verify_save()
        step_index = self.current_step_index + 1
        # Get the lines of a blank step
        self.clear_state()
        step = self.get_step_lines()
        # Insert new step at position step_index
        self.steps.insert(step_index, step)
        # Go to new step
        self.set_step(step_index)
        # Write changes to disk
        self.write_steps_file(self.steps_filename)

    def insert_steps(self, filename, index):
        # Read the steps file
        (steps, _) = self.read_steps_file(filename)

        data_dir = ppm_utils.get_data_dir(filename)
        for step in steps:
            self.set_state_from_step_lines(step, data_dir, warn=False)
            step = self.get_step_lines(self.steps_data_dir)

        # Insert steps into self.steps
        self.steps[index:index] = steps
        # Write changes to disk
        self.write_steps_file(self.steps_filename)

    def insert_steps_before(self, action):
        if not self.steps_filename:
            return
        # Let the user choose a steps file
        current_filename = self.steps_filename
        filename = self.filedialog("Choose steps file",
                                   default_filename=current_filename)
        if not filename:
            return
        self.verify_save()

        step_index = self.current_step_index
        # Insert steps at position step_index
        self.insert_steps(filename, step_index)
        # Go to new steps
        self.set_step(step_index)

    def insert_steps_after(self, action):
        if not self.steps_filename:
            return
        # Let the user choose a steps file
        current_filename = self.steps_filename
        filename = self.filedialog("Choose steps file",
                                   default_filename=current_filename)
        if not filename:
            return
        self.verify_save()

        step_index = self.current_step_index + 1
        # Insert new steps at position step_index
        self.insert_steps(filename, step_index)
        # Go to new steps
        self.set_step(step_index)

    def cleanup(self, action):
        if not self.steps_filename:
            return
        if not self.question_yes_no("All unused PPM files will be moved to a"
                                    " backup directory. Are you sure?",
                                    "Clean up data directory?"):
            return
        # Remember the current step index
        current_step_index = self.current_step_index
        # Get the backup dir
        backup_dir = os.path.join(self.steps_data_dir, "backup")
        # Create it if it doesn't exist
        if not os.path.exists(backup_dir):
            os.makedirs(backup_dir)
        # Move all files to the backup dir
        for filename in glob.glob(os.path.join(self.steps_data_dir,
                                               "*.[Pp][Pp][Mm]")):
            shutil.move(filename, backup_dir)
        # Get the used files back
        for step in self.steps:
            self.set_state_from_step_lines(step, backup_dir, warn=False)
            self.get_step_lines(self.steps_data_dir)
        # Remove the used files from the backup dir
        used_files = os.listdir(self.steps_data_dir)
        for filename in os.listdir(backup_dir):
            if filename in used_files:
                os.unlink(os.path.join(backup_dir, filename))
        # Restore step index
        self.set_step(current_step_index)
        # Inform the user
        self.message("All unused PPM files may be found at %s." %
                     os.path.abspath(backup_dir),
                     "Clean up data directory")

    # Methods

    def read_steps_file(self, filename):
        steps = []
        header = ""

        fileobj = open(filename, "r")
        for line in fileobj.readlines():
            words = line.split()
            if not words:
                continue
            if line.startswith("# ----"):
                continue
            if words[0] == "step":
                steps.append("")
            if steps:
                steps[-1] += line
            else:
                header += line
        fileobj.close()

        return (steps, header)

    def set_steps_file(self, filename):
        try:
            (self.steps, self.header) = self.read_steps_file(filename)
        except (TypeError, IOError):
            self.message("Cannot read file %s." % filename, "Error")
            return

        self.steps_filename = filename
        self.steps_data_dir = ppm_utils.get_data_dir(filename)
        # Go to step 0
        self.set_step(0)

    def set_step(self, index):
        # Limit index to legal boundaries
        if index < 0:
            index = 0
        if index > len(self.steps) - 1:
            index = len(self.steps) - 1

        # Enable the menus
        self.menu_edit.set_sensitive(True)
        self.menu_insert.set_sensitive(True)
        self.menu_tools.set_sensitive(True)

        # If no steps exist...
        if self.steps == []:
            self.current_step_index = index
            self.current_step = None
            # Set window title
            self.window.set_title("Step Editor -- %s" %
                                  os.path.basename(self.steps_filename))
            # Set step entry widget text
            self.entry_step_num.set_text("")
            # Clear the state of all widgets
            self.clear_state()
            # Disable the main vbox
            self.main_vbox.set_sensitive(False)
            return

        self.current_step_index = index
        self.current_step = self.steps[index]
        # Set window title
        self.window.set_title("Step Editor -- %s -- step %d" %
                              (os.path.basename(self.steps_filename),
                               index + 1))
        # Set step entry widget text
        self.entry_step_num.set_text(str(self.current_step_index + 1))
        # Load the state from the step lines
        self.set_state_from_step_lines(self.current_step, self.steps_data_dir)
        # Enable the main vbox
        self.main_vbox.set_sensitive(True)
        # Make sure the step lines in self.current_step are identical to the
        # output of self.get_step_lines
        self.current_step = self.get_step_lines()

    def verify_save(self):
        if not self.steps:
            return
        # See if the user changed anything
        if self.get_step_lines() != self.current_step:
            if self.question_yes_no("Step contents have been modified."
                                    " Save step?", "Save changes?"):
                self.save_step()

    def save_step(self):
        lines = self.get_step_lines(self.steps_data_dir)
        if lines is not None:
            self.steps[self.current_step_index] = lines
            self.current_step = lines
            self.write_steps_file(self.steps_filename)

    def write_steps_file(self, filename):
        fileobj = open(filename, "w")
        fileobj.write(self.header)
        for step in self.steps:
            fileobj.write("# " + "-" * 32 + "\n")
            fileobj.write(step)
        fileobj.close()


if __name__ == "__main__":
    se = StepEditor()
    if len(sys.argv) > 1:
        se.set_steps_file(sys.argv[1])
    gtk.main()



"""
vsock device support class(es)

https://libvirt.org/formatdomain.html#vsock
"""

from virttest.libvirt_xml import accessors
from virttest.libvirt_xml.devices import base, librarian


class Vsock(base.UntypedDeviceBase):

    __slots__ = ('model_type', 'cid', 'address', 'alias')

    def __init__(self, virsh_instance=base.base.virsh):
        accessors.XMLAttribute('model_type', self,
                               parent_xpath='/',
                               tag_name='vsock',
                               attribute='model')
        accessors.XMLElementDict('cid', self,
                                 parent_xpath='/',
                                 tag_name='cid')
        accessors.XMLElementNest('address', self, parent_xpath='/',
                                 tag_name='address', subclass=self.Address,
                                 subclass_dargs={
                                     'type_name': 'pci',
                                     'virsh_instance': virsh_instance})
        accessors.XMLElementDict('alias', self,
                                 parent_xpath='/',
                                 tag_name='alias')
        super(Vsock, self).__init__(device_tag='vsock',
                                    virsh_instance=virsh_instance)
        self.xml = '<vsock/>'

    Address = librarian.get('address')

    def new_vsock_address(self, **dargs):
        """
        Return a new interface Address instance and set properties from dargs
        """
        new_one = self.Address("pci", virsh_instance=self.virsh)
        for key, value in list(dargs.items()):
            setattr(new_one, key, value)
        return new_one



"""
Windows WMIC utilities
"""

import re


_WMIC_CMD = "wmic"

FMT_TYPE_LIST = "/format:list"


def make_query(cmd, cond=None, props=None, get_swch=None, gbl_swch=None):
    """
    Make a WMIC query command. The command pattern is:

      wmic [GBL_SWCH] CMD [where COND] get [PROPS] [GET_SWCH]

    :param cmd: WMIC command.
    :param cond: Query condition to be appended to `where`.
    :param props: Properties to be get.
    :param get_swch: Local switch of `get`.
    :param gbl_swch: Global switch.

    :return: Query command.
    """
    query = [_WMIC_CMD, cmd]
    if gbl_swch:
        query.append(gbl_swch)
    if cond:
        query.append("where (%s)" % cond)
    query.append("get")
    if props:
        query.append(",".join(props))
    if get_swch:
        query.append(get_swch)
    return " ".join(query)


def is_noinstance(data):
    """
    Check if the given WMIC data contains instance(s).

    :param data: The given WMIC data.

    :return: True if no instance(s), else False.
    """
    return bool(re.match(r"No Instance\(s\) Available", data.strip(), re.I))


def parse_list(data):
    """
    Parse the given list format WMIC data.

    :param data: The given WMIC data.

    :return: Formated data.
    """
    out = []
    if not is_noinstance(data):
        for para in re.split("(?:\r?\n){2,}", data.strip()):
            keys, vals = [], []
            for line in para.splitlines():
                key, val = line.split('=', 1)
                keys.append(key)
                vals.append(val)
            if len(keys) == 1:
                out.append(vals[0])
            else:
                out.append(dict(zip(keys, vals)))
    return out



"""
Common base classes for devices
"""

import logging
from six import StringIO

from virttest import xml_utils
from virttest.libvirt_xml import base, xcepts, accessors
from virttest.xml_utils import ElementTree


class UntypedDeviceBase(base.LibvirtXMLBase):

    """
    Base class implementing common functions for all device XML w/o a type attr.
    """

    __slots__ = ('device_tag',)

    # Subclasses are expected to hide device_tag
    def __init__(self, device_tag, virsh_instance=base.virsh):
        """
        Initialize untyped device instance's basic XML with device_tag
        """
        super(UntypedDeviceBase, self).__init__(virsh_instance=virsh_instance)
        # Just a regular dictionary value
        # (Using a property to change element tag won't work)
        self['device_tag'] = device_tag
        # setup bare-bones XML
        self.xml = u"<%s/>" % device_tag

    def from_element(self, element):
        """
        Stateful component to helper method for new_from_element.
        """
        class_name = self.__class__.__name__
        if element.tag != class_name.lower():
            raise xcepts.LibvirtXMLError('Refusing to create %s instance'
                                         'from %s tagged element'
                                         % (class_name, element.tag))
        # XMLTreeFile only supports element trees
        etree = xml_utils.ElementTree.ElementTree(element)
        # ET only writes to open file-like objects
        xmlstr = StringIO()
        # Need element tree string value to initialize LibvirtXMLBase.xml
        etree.write(xmlstr, xml_utils.ENCODING)
        # Create a new XMLTreeFile object based on string input
        self.xml = xmlstr.getvalue()

    @classmethod
    def new_from_element(cls, element, virsh_instance=base.virsh):
        """
        Create a new device XML instance from an single ElementTree element
        """
        # subclasses __init__ only takes virsh_instance parameter
        instance = cls(virsh_instance=virsh_instance)
        instance.from_element(element)
        return instance

    @classmethod
    def new_from_dict(cls, properties, virsh_instance=base.virsh):
        """
        Create a new device XML instance from a dict-like object
        """
        instance = cls(virsh_instance=virsh_instance)
        for key, value in list(properties.items()):
            setattr(instance, key, value)
        return instance

    # Add accessors here to be used by any elements
    def _get_list(self, tag_filter):
        """
        Return a list of dictionaries containing element's attributes.
        """
        dict_list = []
        elements = self.xmltreefile.findall(tag_filter)
        for element in elements:
            dict_list.append(dict(list(element.items())))
        return dict_list

    def _set_list(self, tag_name, value):
        """
        Set all elements to the value list of dictionaries of element's
        attributes.
        """
        xcept = xcepts.LibvirtXMLError("Must set %s child %s elements from"
                                       " a list of dictionary"
                                       % (self.device_tag, tag_name))
        if not isinstance(value, list):
            raise xcept
        # Start with clean slate
        self._del_list(tag_name)
        for dict_item in value:
            if not isinstance(dict_item, dict):
                raise xcept
            ElementTree.SubElement(self.xmltreefile.getroot(),
                                   tag_name, dict_item)
        self.xmltreefile.write()

    def _del_list(self, tag_filter):
        """
        Remove the list of dictionaries containing each element's attributes.
        """
        element = self.xmltreefile.find(tag_filter)
        while element is not None:
            self.xmltreefile.getroot().remove(element)
            element = self.xmltreefile.find(tag_filter)
        self.xmltreefile.write()

    def _add_item(self, prop_name, **attributes):
        """
        Convenience method for appending an element from dictionary of
        attributes.
        """
        items = self[prop_name]  # xml element name
        items.append(attributes)
        self[prop_name] = items

    def _update_item(self, prop_name, index, **attributes):
        """
        Convenience method for merging values into an element's attributes
        """
        items = self[prop_name]  # xml element name
        item = items[index]
        item.update(attributes)
        self[prop_name] = items


class TypedDeviceBase(UntypedDeviceBase):

    """
    Base class implementing common functions for all device XML w/o a type attr.
    """

    __slots__ = ('type_name',)

    # Subclasses are expected to hide device_tag
    def __init__(self, device_tag, type_name, virsh_instance=base.virsh):
        """
        Initialize Typed device instance's basic XML with type_name & device_tag
        """
        # generate getter, setter, deleter for 'type_name' property
        accessors.XMLAttribute('type_name', self,
                               # each device is it's own XML "document"
                               # because python 2.6 ElementPath is broken
                               parent_xpath='/',
                               tag_name=device_tag,
                               attribute='type')
        super(TypedDeviceBase, self).__init__(device_tag=device_tag,
                                              virsh_instance=virsh_instance)
        # Calls accessor to modify xml
        self.type_name = type_name

    @classmethod
    def new_from_element(cls, element, virsh_instance=base.virsh):
        """
        Hides type_name from superclass new_from_element().
        """
        type_name = element.get('type', None)
        # subclasses must hide device_tag parameter
        instance = cls(type_name=type_name,
                       virsh_instance=virsh_instance)
        instance.from_element(element)
        return instance


# Metaclass is a type-of-types or a class-generating class.
# Using it here to avoid copy-pasting very similar class
# definitions into every unwritten device module.
#
# Example usage for stub disk device:
#
# class Disk(base.TypedDeviceBase):
#     __metaclass__ = base.StubDeviceMeta
#     _device_tag = 'disk'
#     _def_type_name = 'block'
#
# will become defined as:
#
# class Disk(base.TypedDeviceBase):
#     def __init__(self, type_name='block', virsh_instance=base.virsh):
#         issue_warning()
#         super(Disk, self).__init__(device_tag='disk'),
#                                    type_name=type_name,
#                                    virsh_instance=virsh_instance)
#

class StubDeviceMeta(type):

    """
    Metaclass for generating stub Device classes where not fully implemented yet
    """

    warning_issued = False

    # mcs is the class object being generated, name is it's name, bases
    # is tuple of all baseclasses, and dct is what will become mcs's
    # __dict__ after super(...).__init__() is called.
    def __init__(mcs, name, bases, dct):
        """
        Configuration for new class
        """

        # Keep pylint happy
        dct = dict(dct)

        # Call type() to setup new class and store it as 'mcs'
        super(StubDeviceMeta, mcs).__init__(name, bases, dct)

        # Needed for UntypedDeviceBase __init__'s default argument value
        # i.e. device_tag='disk' as specified by specific device class
        if not hasattr(mcs, '_device_tag'):
            raise ValueError(
                "Class %s requires a _device_tag attribute" % name)

        # Same message for both TypedDeviceBase & UntypedDeviceBase subclasses
        message = ("Detected use of a stub device XML for a %s class. These "
                   "only implement a minimal interface that is very likely to "
                   "change in future versions.  This warning will only be "
                   " logged once." % name)

        def issue_warning():
            """
            Closure for created __init__ to only print message once.
            """
            # Examine the CLASS variable
            if not StubDeviceMeta.warning_issued:
                # Set the CLASS variable
                StubDeviceMeta.warning_issued = True
                logging.warning(message)
            else:
                pass  # do nothing

        # Create the proper init function for subclass type
        if TypedDeviceBase in bases:
            # Needed for TypedDeviceBase __init__'s default argument value
            # i.e. type_name='pci' as specified by specific device class.
            if not hasattr(mcs, '_def_type_name'):
                raise ValueError("TypedDevice sub-Class %s must define a "
                                 "_def_type_name attribute" % name)
            # form __init__() and it's arguments for generated class

            def stub_init(self, type_name=getattr(mcs, '_def_type_name'),
                          virsh_instance=base.virsh):
                """
                Initialize stub typed device instance
                """
                # issue warning only when some code instantiats
                # object from generated class
                issue_warning()
                # Created class __init__ still needs to call superclass
                # __init__ (i.e. UntypedDeviceBase or TypedDeviceBase)
                TypedDeviceBase.__init__(self, device_tag=getattr(mcs,
                                                                  '_device_tag'),
                                         type_name=type_name,
                                         virsh_instance=virsh_instance)
        elif UntypedDeviceBase in bases:
            # generate __init__() for untyped devices (similar to above)
            def stub_init(self, virsh_instance=base.virsh):
                """
                Initialize stub un-typed device instance
                """
                issue_warning()
                UntypedDeviceBase.__init__(self, device_tag=getattr(mcs,
                                                                    '_device_tag'),
                                           virsh_instance=virsh_instance)
        else:
            # unexpected usage
            raise TypeError("Class %s is not a subclass of TypedDeviceBase or "
                            "UntypedDeviceBase")
        # Point the generated class's __init__ at the generated function above
        setattr(mcs, '__init__', stub_init)



import os
import uuid

from virttest.libvirt_xml.devices.filesystem import Filesystem


def create_fs_xml(fsdev_dict):
    """
    Create filesystem device xml

    :param fsdev_dict: filesystem device paramter dict
    :return: filesystem device
    """
    accessmode = fsdev_dict.get("accessmode", "passthrough")
    driver = fsdev_dict.get("driver", {'type': 'virtiofs'})
    source = fsdev_dict.get("source", {'dir': '/fs_share_dir'})
    target = fsdev_dict.get("target", {'dir': 'fs_mount'})
    binary_dict = fsdev_dict.get('binary', {})

    not os.path.isdir(source['dir']) and os.mkdir(source['dir'])
    fs_dev = Filesystem()
    fs_dev.accessmode = accessmode
    fs_dev.driver = driver
    fs_dev.source = source
    fs_dev.target = target
    fs_dev.alias = {'name': 'ua-' + str(uuid.uuid1())}

    binary_xml = Filesystem.Binary()
    path = binary_dict.get('path', '/usr/libexec/virtiofsd')
    cache_mode = binary_dict.get('cache_mode', 'none')
    xattr = binary_dict.get('xattr', 'on')
    lock_posix = binary_dict.get('lock_posix', 'on')
    flock = binary_dict.get('flock', 'on')
    if cache_mode != "auto":
        binary_xml.cache_mode = cache_mode
    if xattr != "":
        binary_xml.xattr = xattr
    binary_xml.path = path
    binary_xml.lock_posix = lock_posix
    binary_xml.flock = flock
    fs_dev.binary = binary_xml

    return fs_dev



"""
funcatexit.py - allow programmer to define multiple exit functions to be
executed upon normal cases termination. Can be used for the environment clean
up functions. The basic idea is like atexit from python libs.
"""

__all__ = ["register", "run_exitfuncs", "unregister"]

import traceback

from avocado.core import exceptions

import six


def run_exitfuncs(env, test_type):
    """
    Run any registered exit functions.
    exithandlers is traversed in reverse order so functions are executed
    last in, first out.

    param env: the global objects used by tests
    param test_type: test type mark for exit functions
    """
    error_message = ""
    if env.data.get("exithandlers__%s" % test_type):
        exithandlers = env.data.get("exithandlers__%s" % test_type)
        while exithandlers:
            func, targs, kargs = exithandlers.pop()
            try:
                func(*targs, **kargs)
            except Exception as details:
                error_message += "Error in %s:" % func.__name__
                error_message += " %s\n" % details
                traceback.print_exc()

    return error_message


def register(env, test_type, func, *targs, **kargs):
    """
    Register a function to be executed upon case termination.
    func is returned to facilitate usage as a decorator.

    param env: the global objects used by tests
    param test_type: test type mark for exit functions
    param func: function to be called at exit
    param targs: optional arguments to pass to func
    param kargs: optional keyword arguments to pass to func
    """
    # Check for unpickable arguments
    if func.__name__ not in func.__globals__:
        raise exceptions.TestError("Trying to register function '%s', which is not "
                                   "declared at module scope (not in globals). "
                                   "Please contact the test developer to fix it."
                                   % func)
    for arg in targs:
        if hasattr(arg, '__slots__') and not hasattr(arg, '__getstate__'):
            raise exceptions.TestError("Trying to register exitfunction '%s' with "
                                       "unpickable targument '%s'. Please contact "
                                       "the test developer to fix it."
                                       % (func, arg))
    for key, arg in six.iteritems(kargs):
        if hasattr(arg, '__slots__') and not hasattr(arg, '__getstate__'):
            raise exceptions.TestError("Trying to register exitfunction '%s' with "
                                       "unpickable kargument '%s=%s'. Please "
                                       "contact the test developer to fix it."
                                       % (func, key, arg))
    exithandlers = "exithandlers__%s" % test_type
    if not env.data.get(exithandlers):
        env.data[exithandlers] = []

    env.data[exithandlers].append((func, targs, kargs))
    return func


def unregister(env, test_type, func, *targs, **kargs):
    """
    Unregister a function to be executed upon case termination.
    func is returned to facilitate usage as a decorator.

    param env: the global objects used by tests
    param test_type: test type mark for exit functions
    param func: function to be called at exit
    param targs: optional arguments to pass to func
    param kargs: optional keyword arguments to pass to func
    """
    exithandlers = "exithandlers__%s" % test_type
    if env.data.get(exithandlers):
        env.data[exithandlers].remove((func, targs, kargs))
    return func



"""
Classes and functions to handle block/disk images for KVM.

This exports:
  - two functions for get image/blkdebug filename
  - class for image operates and basic parameters
"""
import collections
import json
import logging
import os
import re
import six
import string

from avocado.core import exceptions
from avocado.utils import process

from virttest import utils_misc
from virttest import virt_vm
from virttest import storage
from virttest import nvme
from virttest import data_dir
from virttest import error_context


def filename_to_file_opts(filename):
    """Convert filename into file opts, used by both qemu-img and qemu-kvm"""
    file_opts = {}
    if not filename:
        file_opts = {}
    elif filename.startswith('iscsi:'):
        filename_pattern = re.compile(
            r'iscsi://((?P<user>.+?):(?P<password>.+?)@)?(?P<portal>.+)/(?P<target>.+?)/(?P<lun>\d+)')
        matches = filename_pattern.match(filename)
        if matches:
            if (matches.group('portal') is not None
                    and matches.group('target') is not None
                    and matches.group('lun') is not None):
                # required options for iscsi
                file_opts = {'driver': 'iscsi',
                             'transport': 'tcp',
                             'portal': matches.group('portal'),
                             'target': matches.group('target'),
                             'lun': int(matches.group('lun'))}
                if matches.group('user') is not None:
                    # optional option
                    file_opts['user'] = matches.group('user')
    elif filename.startswith('rbd:'):
        filename_pattern = re.compile(
            r'rbd:(?P<pool>.+?)/(?P<namespace>.+?(?=/))?/?(?P<image>[^:]+)'
            r'(:conf=(?P<conf>.+))?'
        )
        matches = filename_pattern.match(filename)
        if matches:
            if (matches.group('pool') is not None
                    and matches.group('image') is not None):
                # required options for rbd
                file_opts = {'driver': 'rbd',
                             'pool': matches.group('pool'),
                             'image': matches.group('image')}
                if matches.group('conf') is not None:
                    # optional option
                    file_opts['conf'] = matches.group('conf')
                if matches.group('namespace') is not None:
                    # optional option
                    file_opts['namespace'] = matches.group('namespace')
    elif filename.startswith('gluster'):
        filename_pattern = re.compile(
            r'gluster\+?(?P<type>.+)?://((?P<host>[^/]+?)(:(?P<port>\d+))?)?/'
            r'(?P<volume>.+?)/(?P<path>[^,?]+)'
            r'(\?socket=(?P<socket>[^,]+))?'
        )
        matches = filename_pattern.match(filename)
        if matches:
            servers = []
            transport = 'inet' if not matches.group('type') or matches.group('type') == 'tcp' else matches.group('type')

            if matches.group('host'):
                # 'IPv4/hostname' or '[IPv6 address]'
                host = matches.group('host').strip('[]')

                # port should be set for both qemu-img and qemu-kvm
                port = matches.group('port') if matches.group('port') else '0'

                servers.append({'type': transport,
                                'host': host,
                                'port': port})
            elif matches.group('socket'):
                servers.append({'type': transport,
                                'path': matches.group('socket')})

            if matches.group('volume') and matches.group('path') and servers:
                # required options for gluster
                file_opts = {'driver': 'gluster',
                             'volume': matches.group('volume'),
                             'path': matches.group('path')}
                file_opts.update({'server.{i}.{k}'.format(i=i, k=k): v
                                 for i, server in enumerate(servers)
                                 for k, v in six.iteritems(server)})
    elif re.match(r'nbd(\+\w+)?://', filename):
        filename_pattern = re.compile(
            r'nbd(\+(?:.+))?://((?P<host>[^/:?]+)(:(?P<port>\d+))?)?'
            r'(/(?P<export>[^?]+))?'
            r'(\?socket=(?P<socket>.+))?'
        )
        matches = filename_pattern.match(filename)
        if matches:
            server = {}
            host = matches.group('host')
            sock = matches.group('socket')

            if host:
                # 10890 is the default port for tcp connection
                port = matches.group('port') if matches.group(
                                                        'port') else '10809'
                server = {'server.type': 'inet',
                          'server.host': host,
                          'server.port': port}
            elif sock:
                server = {'server.type': 'unix', 'server.path': sock}

            if server:
                # server is required
                file_opts = {'driver': 'nbd'}
                file_opts.update(server)

                if matches.group('export'):
                    file_opts['export'] = matches.group('export')
    elif filename.startswith('nvme:'):
        addr, namespace = nvme.parse_uri(filename)
        file_opts = {'driver': 'nvme', 'device': addr, 'namespace': int(namespace)}
    elif filename.startswith('ssh:'):
        filename_pattern = re.compile(
            r'ssh://((?P<user>.+)@)?(?P<host>[^/:?]+)(:(?P<port>\d+))?'
            r'(?P<path>/[^?]+)'
            r'(\?host_key_check=(?P<host_key_check>.+))?'
        )
        matches = filename_pattern.match(filename)
        if matches:
            matches = matches.groupdict()
            if matches['host'] is not None and matches['path'] is not None:
                # required ssh options
                file_opts = {
                    'driver': 'ssh',
                    'server.host': matches['host'],
                    'server.port': matches['port'] if matches['port'] else 22,
                    'path': matches['path']
                }

                if matches['user'] is not None:
                    file_opts['user'] = matches['user']

                # options in qemu-kvm are different from uri
                if matches['host_key_check'] is not None:
                    if matches['host_key_check'] == 'no':
                        file_opts['host-key-check.mode'] = 'none'
                    elif matches['host_key_check'] == 'yes':
                        file_opts['host-key-check.mode'] = 'known_hosts'
                    else:
                        m = re.match(r'(?P<type>md5|sha1):(?P<hash>.+)',
                                     matches['host_key_check']).groupdict()
                        file_opts.update({
                            'host-key-check.mode': 'hash',
                            'host-key-check.type': m['type'],
                            'host-key-check.hash': m['hash']
                        })
    elif re.match(r'(http|https|ftp|ftps)://', filename):
        filename_pattern = re.compile(
            r'(?P<protocol>.+?)://((?P<user>.+?)(:(?P<password>.+?))?@)?'
            r'(?P<server>.+?)/(?P<path>.+)')
        matches = filename_pattern.match(filename)
        if matches:
            matches = matches.groupdict()
            if all((matches['protocol'], matches['server'], matches['path'])):
                # required libcurl options, note server can be hostname:port
                file_opts = {
                    'driver': matches['protocol'],
                    'url': '{protocol}://{server}/{path}'.format(
                        protocol=matches['protocol'],
                        server=matches['server'],
                        path=matches['path']
                    )
                }

                if matches['user'] is not None:
                    file_opts['username'] = matches['user']
    # FIXME: Judge the host device by the string starts with "/dev/".
    elif filename.startswith('/dev/'):
        file_opts = {'driver': 'host_device', 'filename': filename}
    else:
        file_opts = {'driver': 'file', 'filename': filename}

    if not file_opts:
        raise ValueError("Wrong filename %s" % filename)

    return file_opts


def _get_image_meta(image, params, root_dir):
    """Retrieve image meta dict."""
    meta = collections.OrderedDict()
    meta["file"] = collections.OrderedDict()

    filename = storage.get_image_filename(params, root_dir)
    meta_file = filename_to_file_opts(filename)
    meta["file"].update(meta_file)

    image_format = params.get("image_format", "qcow2")
    meta["driver"] = image_format

    secret = storage.ImageSecret.image_secret_define_by_params(image, params)
    if image_format == "luks":
        meta["key-secret"] = secret.aid
    image_encryption = params.get("image_encryption", "off")
    if image_format == "qcow2" and image_encryption == "luks":
        meta["encrypt.key-secret"] = secret.aid

    auth_info = storage.StorageAuth.auth_info_define_by_params(image, params)
    if auth_info is not None:
        if auth_info.storage_type == 'ceph':
            if auth_info.data:
                # qemu-img needs secret object only for ceph access
                meta['file']['password-secret'] = auth_info.aid
        elif auth_info.storage_type == 'iscsi-direct':
            if auth_info.data:
                # '-b json' demands password
                # note that image creation doesn't support secret object
                meta['file']['password'] = auth_info.data
            if auth_info.iscsi_initiator:
                meta['file']['initiator-name'] = auth_info.iscsi_initiator
        elif auth_info.storage_type == 'glusterfs-direct':
            if auth_info.debug:
                meta['file']['debug'] = int(auth_info.debug)
            if auth_info.logfile:
                meta['file']['logfile'] = auth_info.logfile

            peers = []
            for peer in auth_info.peers:
                if 'path' in peer:
                    # access storage with unix domain socket
                    peers.append({'type': 'unix', 'path': peer['path']})
                else:
                    # access storage with hostname/ip + port
                    peers.append({'host': peer['host'],
                                  'type': peer.get('type', 'inet'),
                                  'port': '%s' % peer.get('port', '0')})
            meta['file'].update({'server.{i}.{k}'.format(i=i+1, k=k): v
                                for i, server in enumerate(peers)
                                for k, v in six.iteritems(server)})
        elif auth_info.storage_type == 'nbd':
            # qemu-img, as a client, accesses nbd storage
            if auth_info.tls_creds:
                meta['file']['tls-creds'] = auth_info.aid
            if auth_info.reconnect_delay:
                meta['file']['reconnect-delay'] = auth_info.reconnect_delay
        elif auth_info.storage_type == 'curl':
            mapping = {
                'password-secret': (auth_info.data, auth_info.aid),
                'sslverify': (auth_info.sslverify, auth_info.sslverify),
                'cookie-secret': (auth_info.cookie,
                                  auth_info.cookie.aid
                                  if auth_info.cookie else ''),
                'readahead': (auth_info.readahead, auth_info.readahead),
                'timeout': (auth_info.timeout, auth_info.timeout)
            }
            meta['file'].update({
                k: v[1] for k, v in six.iteritems(mapping) if v[0]
            })

    return meta


def get_image_json(image, params, root_dir):
    """Generate image json representation."""
    return "json:%s" % json.dumps(_get_image_meta(image, params, root_dir))


def get_image_opts(image, params, root_dir):
    """Generate image-opts."""
    def _dict_to_dot(dct):
        """Convert dictionary to dot representation."""
        flat = []
        prefix = []
        stack = [six.iteritems(dct)]
        while stack:
            it = stack[-1]
            try:
                key, value = next(it)
            except StopIteration:
                if prefix:
                    prefix.pop()
                stack.pop()
                continue
            if isinstance(value, collections.Mapping):
                prefix.append(key)
                stack.append(six.iteritems(value))
            else:
                flat.append((".".join(prefix + [key]), value))
        return flat

    meta = _get_image_meta(image, params, root_dir)
    return ",".join(["%s=%s" % (attr, value) for
                     attr, value in _dict_to_dot(meta)])


def get_image_repr(image, params, root_dir, representation=None):
    """Get image representation."""
    mapping = {"filename": lambda i, p, r: storage.get_image_filename(p, r),
               "json": get_image_json,
               "opts": get_image_opts}
    func = mapping.get(representation, None)
    if func is None:
        image_secret = storage.ImageSecret.image_secret_define_by_params(
                                                           image, params)

        access_needed = False
        auth_info = storage.StorageAuth.auth_info_define_by_params(image,
                                                                   params)
        if auth_info is not None:
            if auth_info.storage_type == 'ceph':
                # only ceph access needs secret object
                if auth_info.data:
                    access_needed = True
            elif auth_info.storage_type == 'iscsi-direct':
                # url with u/p is used to access iscsi image,
                # besides u/p, iscsi access may need initiator
                if auth_info.iscsi_initiator:
                    access_needed = True
            elif auth_info.storage_type == 'glusterfs-direct':
                # debug, logfile and other servers represent in json
                if auth_info.debug or auth_info.logfile or auth_info.peers:
                    access_needed = True
            elif auth_info.storage_type == 'nbd':
                # tls-creds, reconnect_delay represent in json
                access_needed = True
            elif auth_info.storage_type == 'curl':
                # u/p can be included in url, while the others should be
                # represented in json
                if any((auth_info.sslverify, auth_info.cookie,
                        auth_info.readahead, auth_info.timeout)):
                    access_needed = True

        func = mapping["json"] if image_secret or access_needed else mapping["filename"]
    return func(image, params, root_dir)


class _ParameterAssembler(string.Formatter):
    """
    Command line parameter assembler.

    This will automatically prepend parameter if corresponding value is passed
    to the format string.
    """
    sentinal = object()

    def __init__(self, cmd_params=None):
        string.Formatter.__init__(self)
        self.cmd_params = cmd_params or {}

    def format(self, format_string, *args, **kwargs):
        """Remove redundant whitespaces and return format string."""
        ret = string.Formatter.format(self, format_string, *args, **kwargs)
        return re.sub(" +", " ", ret)

    def get_value(self, key, args, kwargs):
        try:
            val = string.Formatter.get_value(self, key, args, kwargs)
        except KeyError:
            if key in self.cmd_params:
                val = None
            else:
                raise
        return (self.cmd_params.get(key, self.sentinal), val)

    def convert_field(self, value, conversion):
        """
        Do conversion on the resulting object.

        supported conversions:
            'b': keep the parameter only if bool(value) is True.
            'v': keep both the parameter and its corresponding value,
                 the default mode.
        """
        if value[0] is self.sentinal:
            return string.Formatter.convert_field(self, value[1], conversion)
        if conversion is None:
            conversion = "v"
        if conversion == "v":
            return "" if value[1] is None else " ".join(value)
        if conversion == "b":
            return value[0] if bool(value[1]) else ""
        raise ValueError("Unknown conversion specifier {}".format(conversion))


class QemuImg(storage.QemuImg):
    """KVM class for handling operations of disk/block images."""
    qemu_img_parameters = {
        "image_format": "-f",
        "backing_file": "-b",
        "backing_format": "-F",
        "unsafe": "-u",
        "options": "-o",
        "secret_object": "",
        "tls_creds_object": "",
        "image_opts": "",
        "check_repair": "-r",
        "output_format": "--output",
        "force_share": "-U",
        "resize_preallocation": "--preallocation",
        "resize_shrink": "--shrink",
        "convert_compressed": "-c",
        "cache_mode": "-t",
        "source_cache_mode": "-T",
        "target_image_format": "-O",
        "convert_sparse_size": "-S",
        "commit_drop": "-d",
        "compare_strict_mode": "-s",
        "compare_second_image_format": "-F"
        }
    create_cmd = ("create {secret_object} {tls_creds_object} {image_format} "
                  "{backing_file} {backing_format} {unsafe!b} {options} "
                  "{image_filename} {image_size}")
    check_cmd = ("check {secret_object} {tls_creds_object} "
                 "{image_opts} {image_format} "
                 "{output_format} {check_repair} {force_share!b} "
                 "{image_filename}")
    convert_cmd = ("convert {secret_object} {tls_creds_object} "
                   "{convert_compressed!b} {skip_target_image_creation} "
                   "{image_format} {cache_mode} {source_cache_mode} "
                   "{target_image_format} {options} {convert_sparse_size} "
                   "{image_filename} {target_image_filename} "
                   "{target_image_opts}")
    commit_cmd = ("commit {secret_object} {image_format} {cache_mode} "
                  "{backing_file} {commit_drop!b} {image_filename}")
    resize_cmd = ("resize {secret_object} {image_opts} {resize_shrink!b} "
                  "{resize_preallocation} {image_filename} {image_size}")
    rebase_cmd = ("rebase {secret_object} {image_format} {cache_mode} "
                  "{source_cache_mode} {unsafe!b} {backing_file} "
                  "{backing_format} {image_filename}")
    dd_cmd = ("dd {secret_object} {tls_creds_object} {image_format} "
              "{target_image_format} {block_size} {count} {skip} "
              "if={image_filename} of={target_image_filename}")
    compare_cmd = ("compare {secret_object} {tls_creds_object} {image_format} "
                   "{compare_second_image_format} {source_cache_mode} "
                   "{compare_strict_mode!b} {force_share!b} "
                   "{image_filename} {compare_second_image_filename}")

    def __init__(self, params, root_dir, tag):
        """
        Init the default value for image object.

        :param params: Dictionary containing the test parameters.
        :param root_dir: Base directory for relative filenames.
        :param tag: Image tag defined in parameter images
        """
        storage.QemuImg.__init__(self, params, root_dir, tag)
        self.image_cmd = utils_misc.get_qemu_img_binary(params)
        q_result = process.run(self.image_cmd + ' -h', ignore_status=True,
                               shell=True, verbose=False)
        self.help_text = q_result.stdout_text
        self.cap_force_share = '-U' in self.help_text
        self._cmd_formatter = _ParameterAssembler(self.qemu_img_parameters)

    def _parse_options(self, params):
        """Build options used for qemu-img amend, create, convert, measure."""
        options_mapping = {
            "preallocated": (None, "preallocation", ("qcow2", "raw", "luks")),
            "image_cluster_size": (None, "cluster_size", ("qcow2",)),
            "lazy_refcounts": (None, "lazy_refcounts", ("qcow2",)),
            "qcow2_compatible": (None, "compat", ("qcow2",))
        }
        image_format = params.get("image_format", "qcow2")
        options = []
        for key, (default, opt_key, support_fmt) in options_mapping.items():
            if image_format in support_fmt:
                value = params.get(key, default)
                if value is not None:
                    options.append("%s=%s" % (opt_key, value))

        if self.encryption_config.key_secret:
            opts = list(self.encryption_config)
            opts.remove("base_key_secrets")
            if image_format == "luks":
                opts.remove("format")
            for opt_key in opts:
                opt_val = getattr(self.encryption_config, opt_key)
                if opt_val:
                    if image_format == "qcow2":
                        opt_key = "encrypt.%s" % opt_key
                    options.append("%s=%s" % (opt_key.replace("_", "-"),
                                              str(opt_val)))

        if self.data_file:
            options.extend(
                ("data_file=%s" % self.data_file.image_filename,
                 "data_file_raw=%s" % params.get("image_data_file_raw", "off")))

        for access_secret, secret_type in self._image_access_secret:
            if secret_type == 'password':
                options.append("password-secret=%s" % access_secret.aid)
            elif secret_type == 'key':
                options.append("key-secret=%s" % access_secret.aid)
            elif secret_type == 'cookie':
                options.append("cookie-secret=%s" % access_secret.aid)

        image_extra_params = params.get("image_extra_params")
        if image_extra_params:
            options.append(image_extra_params.strip(','))
        if params.get("has_backing_file") == "yes":
            backing_param = params.object_params("backing_file")
            backing_file = storage.get_image_filename(backing_param,
                                                      self.root_dir)
            options.append("backing_file=%s" % backing_file)
            backing_fmt = backing_param.get("image_format")
            options.append("backing_fmt=%s" % backing_fmt)
        return options

    def _need_auth_info(self, image=None):
        """
        Check if a specified image's auth info is required.
        qemu-img's 'json:{}' instead of 'filename(uri)' should
        be used when auth info is required.
        The auth info includes sensitive data like password as
        well as other info like iscsi initiator.

        :param image: image name
        :return: True or False
        """
        needed = False

        if self.image_access is not None:
            tag = image if image else self.tag
            if tag == self.tag:
                needed = self.image_access.image_auth is not None
            else:
                needed = tag in self.image_access.image_backing_auth

        return needed

    @property
    def _image_access_tls_creds(self):
        tls_creds = None
        creds = self.image_access.image_auth if self.image_access else None

        if creds is not None:
            if creds.storage_type == 'nbd':
                if creds.tls_creds:
                    tls_creds = creds

        return tls_creds

    @property
    def _backing_access_tls_creds(self):
        tls_creds_list = []
        creds_list = self.image_access.image_backing_auth.values() if self.image_access else []

        for creds in creds_list:
            if creds.storage_type == 'nbd':
                if creds.tls_creds:
                    tls_creds_list.append(creds)

        return tls_creds_list

    @property
    def _image_access_secret(self):
        """
        Get the access secret object and its type of the image itself,
        the type can be 'key' or 'password' or 'cookie'

        :return: a list of tuple(StorageAuth object, secret type) or []
        :note: an image can have more than one secret objects, e.g.
               access secret object and cookie secret object for libcurl
        """
        secrets = []
        auth = self.image_access.image_auth if self.image_access else None

        if auth is not None:
            if auth.storage_type == 'ceph':
                # ceph image access requires secret object by
                # qemu-img and only 'password-secret' is supported
                if auth.data:
                    secrets.append((auth, 'password'))
            elif auth.storage_type == 'curl':
                # a libcurl image can have more than one secret object
                if auth.data:
                    secrets.append((auth, 'password'))
                if auth.cookie:
                    secrets.append((auth.cookie, 'cookie'))

        return secrets

    @property
    def _backing_access_secrets(self):
        """
        Get the backing images' access secret objects and types,
        the type can be 'key' or 'password'

        :return: a list of (StorageAuth object, secret type) or []
        """
        secrets = []
        info = self.image_access.image_backing_auth.values() if self.image_access else []

        for auth in info:
            if auth.storage_type == 'ceph':
                # ceph image access requires secret object by
                # qemu-img and only 'password-secret' is supported
                if auth.data:
                    secrets.append((auth, 'password'))
            elif auth.storage_type == 'curl':
                if auth.data:
                    secrets.append((auth, 'password'))
                if auth.cookie:
                    secrets.append((auth.cookie, 'cookie'))

        return secrets

    @property
    def _secret_objects(self):
        """All secret objects str needed for command line."""
        secret_objects = self.encryption_config.image_key_secrets
        secret_obj_str = "--object secret,id={s.aid},data={s.data}"
        return [secret_obj_str.format(s=s) for s in secret_objects]

    @property
    def _image_access_tls_creds_object(self):
        """Get the tls-creds object str of the image itself."""
        tls_obj_str = '--object tls-creds-x509,id={s.aid},endpoint=client,dir={s.tls_creds}'
        creds = self._image_access_tls_creds
        return tls_obj_str.format(s=creds) if creds else ''

    @property
    def _backing_access_tls_creds_objects(self):
        """Get all tls-creds object str of the backing images."""
        tls_creds = []
        tls_obj_str = '--object tls-creds-x509,id={s.aid},endpoint=client,dir={s.tls_creds}'

        for creds in self._backing_access_tls_creds:
            tls_creds.append(tls_obj_str.format(s=creds))

        return tls_creds

    @property
    def _image_access_secret_object(self):
        """Get the secret object str of the image itself."""
        secrets = []

        for access_secret, secret_type in self._image_access_secret:
            secret_obj_str = ''
            if secret_type == 'password':
                secret_obj_str = '--object secret,id={s.aid},format={s.data_format},file={s.filename}'
            elif secret_type == 'key' or secret_type == 'cookie':
                secret_obj_str = '--object secret,id={s.aid},format={s.data_format},data={s.data}'
            secrets.append(secret_obj_str.format(s=access_secret))

        return secrets

    @property
    def _backing_access_secret_objects(self):
        """Get all secret object str of the backing images."""
        secrets = []

        for access_secret, secret_type in self._backing_access_secrets:
            secret_obj_str = ''
            if secret_type == 'password':
                secret_obj_str = "--object secret,id={s.aid},format={s.data_format},file={s.filename}"
            elif secret_type == 'key' or secret_type == 'cookie':
                secret_obj_str = "--object secret,id={s.aid},format={s.data_format},data={s.data}"
            secrets.append(secret_obj_str.format(s=access_secret))

        return secrets

    @error_context.context_aware
    def create(self, params, ignore_errors=False):
        """
        Create an image using qemu_img or dd.

        :param params: Dictionary containing the test parameters.
        :param ignore_errors: Whether to ignore errors on the image creation
                              cmd.

        :note: params should contain:

               image_name
                   name of the image file, without extension
               image_format
                   format of the image (qcow2, raw etc)
               image_cluster_size (optional)
                   cluster size for the image
               image_size
                   requested size of the image (a string qemu-img can
                   understand, such as '10G')
               create_with_dd
                   use dd to create the image (raw format only)
               base_image(optional)
                   the base image name when create snapshot
               base_format(optional)
                   the format of base image
               encrypted(optional)
                   if the image is encrypted, allowed values: on and off.
                   Default is "off"
               preallocated(optional)
                   if preallocation when create image, allowed values: off,
                   metadata. Default is "off"

        :return: tuple (path to the image created, process.CmdResult object
                 containing the result of the creation command).
        """
        if params.get(
                "create_with_dd") == "yes" and self.image_format == "raw":
            # maps K,M,G,T => (count, bs)
            human = {'K': (1, 1),
                     'M': (1, 1024),
                     'G': (1024, 1024),
                     'T': (1024, 1048576),
                     }
            if self.size[-1] in human:
                block_size = human[self.size[-1]][1]
                size = int(self.size[:-1]) * human[self.size[-1]][0]
            qemu_img_cmd = ("dd if=/dev/zero of=%s count=%s bs=%sK"
                            % (self.image_filename, size, block_size))
        else:
            cmd_dict = {}
            cmd_dict["image_format"] = self.image_format
            if self.base_tag:
                # if base image has secret, use json representation
                base_key_secrets = self.encryption_config.base_key_secrets
                if (self.base_tag in [s.image_id for s in base_key_secrets]
                        or self._need_auth_info(self.base_tag)):
                    base_params = params.object_params(self.base_tag)
                    cmd_dict["backing_file"] = "'%s'" % \
                        get_image_json(self.base_tag, base_params,
                                       self.root_dir)
                else:
                    cmd_dict["backing_file"] = self.base_image_filename
                    if self.base_format:
                        cmd_dict["backing_format"] = self.base_format

            # secret objects of the backing images
            secret_objects = self._backing_access_secret_objects

            # secret object of the image itself
            if self._image_access_secret_object:
                secret_objects.extend(self._image_access_secret_object)

            image_secret_objects = self._secret_objects
            if image_secret_objects:
                secret_objects.extend(image_secret_objects)
            if secret_objects:
                cmd_dict["secret_object"] = " ".join(secret_objects)

            # tls creds objects of the backing images of the source
            tls_creds_objects = self._backing_access_tls_creds_objects

            # tls creds object of the source image itself
            if self._image_access_tls_creds_object:
                tls_creds_objects.append(self._image_access_tls_creds_object)

            if tls_creds_objects:
                cmd_dict["tls_creds_object"] = " ".join(tls_creds_objects)

            cmd_dict["image_filename"] = self.image_filename
            cmd_dict["image_size"] = self.size
            options = self._parse_options(params)
            if options:
                cmd_dict["options"] = ",".join(options)
            qemu_img_cmd = self.image_cmd + " " + \
                self._cmd_formatter.format(self.create_cmd, **cmd_dict)

        if (params.get("image_backend", "filesystem") == "filesystem"):
            image_dirname = os.path.dirname(self.image_filename)
            if image_dirname and not os.path.isdir(image_dirname):
                e_msg = ("Parent directory of the image file %s does "
                         "not exist" % self.image_filename)
                logging.error(e_msg)
                logging.error("This usually means a serious setup exceptions.")
                logging.error("Please verify if your data dir contains the "
                              "expected directory structure")
                logging.error("Backing data dir: %s",
                              data_dir.get_backing_data_dir())
                logging.error("Directory structure:")
                for root, _, _ in os.walk(data_dir.get_backing_data_dir()):
                    logging.error(root)

                logging.warning("We'll try to proceed by creating the dir. "
                                "Other errors may ensue")
                os.makedirs(image_dirname)

        msg = "Create image by command: %s" % qemu_img_cmd
        error_context.context(msg, logging.info)
        cmd_result = process.run(
            qemu_img_cmd, shell=True, verbose=False, ignore_status=True)
        if cmd_result.exit_status != 0 and not ignore_errors:
            raise exceptions.TestError("Failed to create image %s\n%s" %
                                       (self.image_filename, cmd_result))
        if self.encryption_config.key_secret:
            self.encryption_config.key_secret.save_to_file()
        cmd_result.stdout = cmd_result.stdout_text
        cmd_result.stderr = cmd_result.stderr_text
        return self.image_filename, cmd_result

    def convert(self, params, root_dir, cache_mode=None,
                source_cache_mode=None, skip_target_creation=False):
        """
        Convert image

        :param params: dictionary containing the test parameters
        :param root_dir: dir for save the convert image
        :param cache_mode: The cache mode used to write the output disk image.
                           Valid options are: ``none``, ``writeback``
                           (default), ``writethrough``, ``directsync`` and
                           ``unsafe``.
        :param source_cache_mode: the cache mode used with source image file
        :param skip_target_creation: qemu-img skips the creation of the target
                                     volume if True(-n), i.e. the target image
                                     should be created before image convert
        :note: params should contain:
            convert_target
                the convert target image tag
            compressed
                indicates that target image must be compressed
            sparse_size
                indicate the consecutive number of bytes contains zeros to
                create sparse image during conversion
        """
        convert_target = params["convert_target"]
        convert_params = params.object_params(convert_target)
        convert_image = QemuImg(convert_params, root_dir, convert_target)

        convert_compressed = convert_params.get("convert_compressed")
        sparse_size = convert_params.get("sparse_size")

        cmd_dict = {
            "convert_compressed": convert_compressed == "yes",
            "convert_sparse_size": sparse_size,
            "image_filename": self.image_filename,
            "image_format": self.image_format,
            "target_image_format": convert_image.image_format,
            "target_image_filename": convert_image.image_filename,
            "cache_mode": cache_mode,
            "source_cache_mode": source_cache_mode,
            "skip_target_image_creation": "-n" if skip_target_creation else "",
            "target_image_opts": ""
        }

        options = convert_image._parse_options(convert_params)
        if options:
            cmd_dict["options"] = ",".join(options)

        if skip_target_creation:
            # -o has no effect when skipping image creation
            # This will become an error in future QEMU versions
            if options:
                cmd_dict.pop("options")

            cmd_dict.pop("target_image_format")
            cmd_dict["target_image_filename"] = ""
            cmd_dict["target_image_opts"] = ("--target-image-opts '%s'"
                                             % get_image_opts(
                                                 convert_image.tag,
                                                 convert_image.params,
                                                 convert_image.root_dir))

        if (self.encryption_config.key_secret
                or self._need_auth_info(self.tag)):
            cmd_dict["image_filename"] = "'%s'" % get_image_json(
                self.tag, self.params, self.root_dir)
            cmd_dict.pop("image_format")

        # source images secrets(luks)
        secret_objects = self._secret_objects

        # secret objects of the backing images of the source
        if self._backing_access_secret_objects:
            secret_objects.extend(self._backing_access_secret_objects)

        # secret object of the source image itself
        if self._image_access_secret_object:
            secret_objects.extend(self._image_access_secret_object)

        # target image access secret object
        # target image to be converted never has backing images
        if convert_image._image_access_secret_object:
            secret_objects.extend(convert_image._image_access_secret_object)

        # target image secret(luks)
        if convert_image.encryption_config.key_secret:
            secret_objects.extend(convert_image._secret_objects)

        if secret_objects:
            cmd_dict["secret_object"] = " ".join(secret_objects)

        # tls creds objects of the backing images of the source
        tls_creds_objects = self._backing_access_tls_creds_objects

        # tls creds object of the source image itself
        if self._image_access_tls_creds_object:
            tls_creds_objects.append(self._image_access_tls_creds_object)

        # tls creds object of the target image
        if convert_image._image_access_tls_creds_object:
            tls_creds_objects.append(
                convert_image._image_access_tls_creds_object)

        if tls_creds_objects:
            cmd_dict["tls_creds_object"] = " ".join(tls_creds_objects)

        convert_cmd = self.image_cmd + " " + \
            self._cmd_formatter.format(self.convert_cmd, **cmd_dict)

        logging.info("Convert image %s from %s to %s", self.image_filename,
                     self.image_format, convert_image.image_format)
        process.run(convert_cmd)
        if convert_image.encryption_config.key_secret:
            convert_image.encryption_config.key_secret.save_to_file()

        return convert_target

    def rebase(self, params, cache_mode=None, source_cache_mode=None):
        """
        Rebase image.

        :param params: dictionary containing the test parameters
        :param cache_mode: the cache mode used to write the output disk image,
                           the valid options are: 'none', 'writeback' (default),
                           'writethrough', 'directsync' and 'unsafe'.
        """
        self.check_option("base_image_filename")
        self.check_option("base_format")

        rebase_mode = params.get("rebase_mode")
        cmd_dict = {"image_format": self.image_format,
                    "image_filename": self.image_filename,
                    "cache_mode": cache_mode,
                    "source_cache_mode": source_cache_mode,
                    "unsafe": rebase_mode == "unsafe"}
        secret_objects = self._secret_objects
        if secret_objects:
            cmd_dict["secret_object"] = " ".join(secret_objects)
        if self.encryption_config.key_secret:
            cmd_dict["image_filename"] = "'%s'" % get_image_json(
                self.tag, self.params, self.root_dir)
            cmd_dict.pop("image_format")
        if self.base_tag:
            if self.base_tag == "null":
                cmd_dict["backing_file"] = "''"
            else:
                base_params = self.params.object_params(self.base_tag)
                base_image = QemuImg(base_params, self.root_dir, self.base_tag)
                self.base_image_filename = base_image.image_filename
                self.base_format = base_image.image_format
                if base_image.encryption_config.key_secret:
                    cmd_dict["backing_file"] = "'%s'" % get_image_json(
                        base_image.tag, base_image.params, base_image.root_dir)
                else:
                    cmd_dict["backing_file"] = base_image.image_filename
                    cmd_dict["backing_format"] = base_image.image_format
        else:
            raise exceptions.TestError("Can not find the image parameters need"
                                       " for rebase.")

        logging.info("Rebase snapshot %s to %s..." % (self.image_filename,
                                                      self.base_image_filename))
        rebase_cmd = self.image_cmd + " " + \
            self._cmd_formatter.format(self.rebase_cmd, **cmd_dict)
        process.run(rebase_cmd)

        return self.base_tag

    def commit(self, params={}, cache_mode=None, base=None, drop=False):
        """
        Commit image to it's base file

        :param cache_mode: the cache mode used to write the output disk image,
            the valid options are: 'none', 'writeback' (default),
            'writethrough', 'directsync' and 'unsafe'.
        :param base: the backing file into which the changes will be committed
        :param drop: drop image after commit
        """
        cmd_dict = {"image_format": self.image_format,
                    "image_filename": self.image_filename,
                    "cache_mode": cache_mode, "commit_drop": drop}
        secret_objects = self._secret_objects
        if secret_objects:
            cmd_dict["secret_object"] = " ".join(secret_objects)
        if base:
            base_params = self.params.object_params(base)
            base_image = QemuImg(base_params, self.root_dir, base)
            if base_image.encryption_config.key_secret:
                cmd_dict["backing_file"] = "'%s'" % get_image_json(
                    base, base_params, self.root_dir)
            else:
                cmd_dict["backing_file"] = base_image.image_filename
        if self.encryption_config.key_secret:
            cmd_dict["image_filename"] = "'%s'" % get_image_json(
                self.tag, self.params, self.root_dir)
            cmd_dict.pop("image_format")
        commit_cmd = self.image_cmd + " " + \
            self._cmd_formatter.format(self.commit_cmd, **cmd_dict)
        logging.info("Commit image %s" % self.image_filename)
        process.run(commit_cmd)

        return self.image_filename

    def snapshot_create(self):
        """
        Create a snapshot image.

        :note: params should contain:
               snapshot_image_name -- the name of snapshot image file
        """

        cmd = self.image_cmd
        if self.snapshot_tag:
            cmd += " snapshot -c %s" % self.snapshot_image_filename
        else:
            raise exceptions.TestError("Can not find the snapshot image"
                                       " parameters")
        cmd += " %s" % self.image_filename

        process.run(cmd)

        return self.snapshot_tag

    def snapshot_del(self, blkdebug_cfg=""):
        """
        Delete a snapshot image.

        :param blkdebug_cfg: The configure file of blkdebug

        :note: params should contain:
               snapshot_image_name -- the name of snapshot image file
        """

        cmd = self.image_cmd
        if self.snapshot_tag:
            cmd += " snapshot -d %s" % self.snapshot_image_filename
        else:
            raise exceptions.TestError("Can not find the snapshot image"
                                       " parameters")
        if blkdebug_cfg:
            cmd += " blkdebug:%s:%s" % (blkdebug_cfg, self.image_filename)
        else:
            cmd += " %s" % self.image_filename

        process.run(cmd)

    def snapshot_list(self):
        """
        List all snapshots in the given image
        """
        cmd = self.image_cmd
        cmd += " snapshot -l %s" % self.image_filename

        return process.run(cmd).stdout_text

    def snapshot_apply(self):
        """
        Apply a snapshot image.

        :note: params should contain:
               snapshot_image_name -- the name of snapshot image file
        """
        cmd = self.image_cmd
        if self.snapshot_tag:
            cmd += " snapshot -a %s %s" % (self.snapshot_image_filename,
                                           self.image_filename)
        else:
            raise exceptions.TestError("Can not find the snapshot image"
                                       " parameters")

        process.run(cmd)

    def remove(self):
        """
        Remove an image file.
        """
        logging.debug("Removing image file %s", self.image_filename)
        storage.file_remove(self.params, self.image_filename)

        if self.data_file:
            logging.debug("Removing external data file of image %s",
                          self.data_file.image_filename)
            storage.file_remove(self.data_file.params,
                                self.data_file.image_filename)

        secret_files = []
        if self.encryption_config.key_secret:
            secret_files.append(self.encryption_config.key_secret.filename)

        if self.image_access:
            secrets = []

            # image secret
            if self.image_access.image_auth:
                secrets.append(self.image_access.image_auth)

            # backing secrets
            secrets.extend(self.image_access.image_backing_auth.values())

            for auth in secrets:
                if auth.data:
                    secret_files.append(auth.filename)

        for f in secret_files:
            if os.path.exists(f):
                os.unlink(f)

    def info(self, force_share=False, output="human"):
        """
        Run qemu-img info command on image file and return its output.

        :param output: string of output format(`human`, `json`)
        """
        logging.debug("Run qemu-img info command on %s", self.image_filename)
        backing_chain = self.params.get("backing_chain")
        force_share &= self.cap_force_share
        cmd = self.image_cmd
        cmd += " info"

        if self._image_access_secret_object:
            # secret object of the image itself
            cmd += " %s" % " ".join(self._image_access_secret_object)

        if self._image_access_tls_creds_object:
            # tls creds object of the image itself
            cmd += " %s" % self._image_access_tls_creds_object

        if backing_chain == "yes":
            if self._backing_access_secret_objects:
                # secret objects of the backing images
                cmd += " %s" % " ".join(self._backing_access_secret_objects)

            if self._backing_access_tls_creds_objects:
                # tls creds objects of the backing images
                cmd += " %s" % " ".join(self._backing_access_tls_creds_objects)

            if "--backing-chain" in self.help_text:
                cmd += " --backing-chain"
            else:
                logging.warn("'--backing-chain' option is not supported")

        if force_share:
            cmd += " -U"

        image_filename = self.image_filename
        if self._need_auth_info(self.tag):
            # use json repr when access info is required
            image_filename = "'%s'" % get_image_json(self.tag, self.params,
                                                     self.root_dir)
        if os.path.exists(image_filename) or self.is_remote_image():
            cmd += " %s --output=%s" % (image_filename, output)
            output = process.run(cmd, verbose=True).stdout_text
        else:
            logging.debug("Image file %s not found", image_filename)
            output = None
        return output

    def get_format(self):
        """
        Get the fimage file format.
        """
        image_info = self.info()
        if image_info:
            image_format = re.findall("file format: (\w+)", image_info)[0]
        else:
            image_format = None
        return image_format

    def support_cmd(self, cmd):
        """
        Verifies whether qemu-img supports command cmd.

        :param cmd: Command string.
        """
        supports_cmd = True

        if cmd not in self.help_text:
            logging.error("%s does not support command '%s'", self.image_cmd,
                          cmd)
            supports_cmd = False

        return supports_cmd

    def compare_images(self, image1, image2, strict_mode=False,
                       verbose=True, force_share=False):
        """
        Compare 2 images using the appropriate tools for each virt backend.

        :param image1: image path of first image
        :param image2: image path of second image
        :param strict_mode: Boolean value, True for strict mode,
                            False for default mode.
        :param verbose: Record output in debug file or not

        :return: process.CmdResult object containing the result of the command
        """
        compare_images = self.support_cmd("compare")
        force_share &= self.cap_force_share
        if not compare_images:
            logging.warn("sub-command compare not supported by qemu-img")
            return None
        else:
            logging.info("Comparing images %s and %s", image1, image2)
            compare_cmd = "%s compare" % self.image_cmd
            if force_share:
                compare_cmd += " -U"
            if strict_mode:
                compare_cmd += " -s"
            compare_cmd += " %s %s" % (image1, image2)
            cmd_result = process.run(compare_cmd, ignore_status=True,
                                     shell=True)

            if verbose:
                logging.debug("Output from command: %s",
                              cmd_result.stdout_text)

            if cmd_result.exit_status == 0:
                logging.info("Compared images are equal")
            elif cmd_result.exit_status == 1:
                raise exceptions.TestFail("Compared images differ")
            else:
                raise exceptions.TestError("Error in image comparison")

            return cmd_result

    def compare_to(self, target_image, source_cache_mode=None,
                   strict_mode=False, force_share=False, verbose=True):
        """
        Compare to target image.

        :param target_image: target image object
        :param source_cache_mode: source cache used to open source image
        :param strict_mode: compare fails on sector allocation or image size
        :param force_share: open image in shared mode
        :return: compare result [process.CmdResult]
        """
        if not self.support_cmd("compare"):
            logging.warn("qemu-img subcommand compare not supported")
            return
        force_share &= self.cap_force_share
        logging.info("compare image %s to image %s",
                     self.image_filename, target_image.image_filename)

        cmd_dict = {
            "image_format": self.image_format,
            "compare_second_image_format": target_image.image_format,
            "source_cache_mode": source_cache_mode,
            "compare_strict_mode": strict_mode,
            "force_share": force_share,
            "image_filename": self.image_filename,
            "compare_second_image_filename": target_image.image_filename,
        }

        secret_objects = self._secret_objects + target_image._secret_objects

        # source image's backing access secret objects
        if self._backing_access_secret_objects:
            secret_objects.extend(self._backing_access_secret_objects)

        # source image access secret object
        if self._image_access_secret_object:
            secret_objects.extend(self._image_access_secret_object)

        # target image's backing access secret objects
        if target_image._backing_access_secret_objects:
            secret_objects.extend(target_image._backing_access_secret_objects)

        # target image access secret object
        if target_image._image_access_secret_object:
            secret_objects.extend(target_image._image_access_secret_object)

        # if compared images are in the same snapshot chain,
        # needs to remove duplicated secrets
        secret_objects = list(set(secret_objects))
        cmd_dict["secret_object"] = " ".join(secret_objects)

        # tls creds objects of the backing images of the source
        tls_creds_objects = self._backing_access_tls_creds_objects

        # tls creds object of the source image
        if self._image_access_tls_creds_object:
            tls_creds_objects.append(self._image_access_tls_creds_object)

        # tls creds objects of the backing images of the target
        if target_image._backing_access_tls_creds_objects:
            tls_creds_objects.extend(
                target_image._backing_access_tls_creds_objects)

        # tls creds object of the target image
        if target_image._image_access_tls_creds_object:
            tls_creds_objects.append(
                target_image._image_access_tls_creds_object)

        tls_creds_objects = list(set(tls_creds_objects))
        cmd_dict["tls_creds_object"] = " ".join(tls_creds_objects)

        if (self.encryption_config.key_secret
                or self._need_auth_info(self.tag)):
            cmd_dict["image_filename"] = "'%s'" % \
                get_image_json(self.tag, self.params, self.root_dir)

        if (target_image.encryption_config.key_secret
                or target_image._need_auth_info(target_image.tag)):
            cmd_dict["compare_second_image_filename"] = "'%s'" % \
                get_image_json(target_image.tag, target_image.params,
                               target_image.root_dir)

        compare_cmd = self.image_cmd + " " + \
            self._cmd_formatter.format(self.compare_cmd, **cmd_dict)
        result = process.run(compare_cmd, ignore_status=True, shell=True)

        if verbose:
            logging.debug("compare output:\n%s", result.stdout_text)

        return result

    def check(self, params, root_dir, force_share=False, output=None):
        """
        Check an image using the appropriate tools for each virt backend.

        :param params: Dictionary containing the test parameters.
        :param root_dir: Base directory for relative filenames.
        :param output: The format of the output(json, human).

        :note: params should contain:
               image_name -- the name of the image file, without extension
               image_format -- the format of the image (qcow2, raw etc)

        :return: The output of check result if the image exists, or None.
        """
        image_filename = self.image_filename
        logging.debug("Checking image file %s", image_filename)
        force_share &= self.cap_force_share

        cmd_dict = {"image_filename": image_filename,
                    "force_share": force_share,
                    "output_format": output}
        if (self.encryption_config.key_secret
                or self._need_auth_info(self.tag)):
            cmd_dict["image_filename"] = "'%s'" % get_image_json(
                self.tag, params, root_dir)

        # access secret objects of the backing images
        secret_objects = self._backing_access_secret_objects

        # access secret object of the image itself
        if self._image_access_secret_object:
            secret_objects.extend(self._image_access_secret_object)

        # image(e.g. luks image) secret objects
        image_secret_objects = self._secret_objects
        if image_secret_objects:
            secret_objects.extend(image_secret_objects)

        if secret_objects:
            cmd_dict["secret_object"] = " ".join(secret_objects)

        # tls creds objects of the backing images
        tls_creds_objects = self._backing_access_tls_creds_objects

        # tls creds object of the image itself
        if self._image_access_tls_creds_object:
            tls_creds_objects.append(self._image_access_tls_creds_object)

        if tls_creds_objects:
            cmd_dict["tls_creds_object"] = " ".join(tls_creds_objects)

        check_cmd = self.image_cmd + " " + self._cmd_formatter.format(
            self.check_cmd, **cmd_dict)
        cmd_result = process.run(check_cmd, ignore_status=True,
                                 shell=True, verbose=False)

        return cmd_result

    def check_image(self, params, root_dir, force_share=False):
        """
        Check an image using the appropriate tools for each virt backend.

        :param params: Dictionary containing the test parameters.
        :param root_dir: Base directory for relative filenames.

        :note: params should contain:
               image_name -- the name of the image file, without extension
               image_format -- the format of the image (qcow2, raw etc)

        :raise VMImageCheckError: In case qemu-img check fails on the image.
        """
        image_filename = self.image_filename
        logging.debug("Checking image file %s", image_filename)
        image_is_checkable = self.image_format in ['qcow2', 'qed']
        force_share &= self.cap_force_share

        if (storage.file_exists(params, image_filename) or
                self.is_remote_image()) and image_is_checkable:
            try:
                # FIXME: do we really need it?
                self.info(force_share)
            except process.CmdError:
                logging.error("Error getting info from image %s",
                              image_filename)
            cmd_result = self.check(params, root_dir, force_share)
            # Error check, large chances of a non-fatal problem.
            # There are chances that bad data was skipped though
            if cmd_result.exit_status == 1:
                stdout = cmd_result.stdout_text
                for e_line in stdout.splitlines():
                    logging.error("[stdout] %s", e_line)
                stderr = cmd_result.stderr_text
                for e_line in stderr.splitlines():
                    logging.error("[stderr] %s", e_line)
                chk = params.get("backup_image_on_check_error", "no")
                if chk == "yes":
                    self.backup_image(params, root_dir, "backup", False)
                raise exceptions.TestWarn(
                    "qemu-img check not completed because of internal "
                    "errors. Some bad data in the image may have gone "
                    "unnoticed (%s)" % image_filename)
            # Exit status 2 is data corruption for sure,
            # so fail the test
            elif cmd_result.exit_status == 2:
                stdout = cmd_result.stdout_text
                for e_line in stdout.splitlines():
                    logging.error("[stdout] %s", e_line)
                stderr = cmd_result.stderr_text
                for e_line in stderr.splitlines():
                    logging.error("[stderr] %s", e_line)
                chk = params.get("backup_image_on_check_error", "no")
                if chk == "yes":
                    self.backup_image(params, root_dir, "backup", False)
                raise virt_vm.VMImageCheckError(image_filename)
            # Leaked clusters, they are known to be harmless to data
            # integrity
            elif cmd_result.exit_status == 3:
                raise exceptions.TestWarn("Leaked clusters were noticed"
                                          " during image check. No data "
                                          "integrity problem was found "
                                          "though. (%s)" % image_filename)
        else:
            if not storage.file_exists(params, image_filename):
                logging.debug("Image file %s not found, skipping check",
                              image_filename)
            elif not image_is_checkable:
                logging.debug(
                    "Image format %s is not checkable, skipping check",
                    self.image_format)

    def amend(self, params, cache_mode=None, ignore_status=False):
        """
        Amend the image format specific options for the image

        :param params: dictionary containing the test parameters
        :param cache_mode: the cache mode used to write the output disk image,
                           the valid options are: 'none', 'writeback'
                           (default), 'writethrough', 'directsync' and
                           'unsafe'.
        :param ignore_status: Whether to raise an exception when command
                              returns =! 0 (False), or not (True).

        :note: params may contain amend options:

               amend_size
                   virtual disk size of the image (a string qemu-img can
                   understand, such as '10G')
               amend_compat
                   compatibility level (0.10 or 1.1)
               amend_backing_file
                   file name of a base image
               amend_backing_fmt
                   image format of the base image
               amend_encryption
                   encrypt the image, allowed values: on and off.
                   Default is "off"
               amend_cluster_size
                   cluster size for the image
               amend_preallocation
                   preallocation mode when create image, allowed values: off,
                   metadata. Default is "off"
               amend_lazy_refcounts
                   postpone refcount updates, allowed values: on and off.
                   Default is "off"
               amend_refcount_bits
                   width of a reference count entry in bits
               amend_extra_params
                   additional options, used for extending amend

        :return: process.CmdResult object containing the result of the
                command
        """
        cmd_list = [self.image_cmd, 'amend']
        options = ["%s=%s" % (key[6:], val) for key, val in six.iteritems(params)
                   if key.startswith('amend_')]
        if cache_mode:
            cmd_list.append("-t %s" % cache_mode)
        if options:
            cmd_list.append("-o %s" %
                            ",".join(options).replace("extra_params=", ""))
        cmd_list.append("-f %s %s" % (self.image_format, self.image_filename))
        logging.info("Amend image %s" % self.image_filename)
        cmd_result = process.run(" ".join(cmd_list), ignore_status=False)
        return cmd_result

    def resize(self, size, shrink=False, preallocation=None):
        """
        Qemu image resize wrapper.

        :param size: string of size representations.(eg. +1G, -1k, 1T)
        :param shrink: boolean
        :param preallocation: preallocation mode
        :return: process.CmdResult object containing the result of the
                 command
        """
        cmd_dict = {
            "resize_shrink": shrink,
            "resize_preallocation": preallocation,
            "image_filename": self.image_filename,
            "image_size": size,
            }
        if self.encryption_config.key_secret:
            cmd_dict["image_filename"] = "'%s'" % get_image_json(
                self.tag, self.params, self.root_dir)
        secret_objects = self._secret_objects
        if secret_objects:
            cmd_dict["secret_object"] = " ".join(secret_objects)
        resize_cmd = self.image_cmd + " " + \
            self._cmd_formatter.format(self.resize_cmd, **cmd_dict)
        cmd_result = process.run(resize_cmd, ignore_status=True)
        return cmd_result

    def map(self, output="human"):
        """
        Qemu image map wrapper.

        :param output: string, the map command output format(`human`, `json`)
        :return: process.CmdResult object containing the result of the
                 command
        """
        cmd_list = [self.image_cmd, "map",
                    ("--output=%s" % output), self.image_filename]
        cmd_result = process.run(" ".join(cmd_list), ignore_status=True)
        return cmd_result

    def measure(self, target_fmt, size=None, output="human"):
        """
        Qemu image measure wrapper.

        :param target_fmt: string, the target image format
        :param size: string, the benchmark size of a target_fmt, if `None` it
                     will measure the image object itself with target_fmt
        :param output: string, the measure command output format
                       (`human`, `json`)
        :return: process.CmdResult object containing the result of the
                 command
        """
        cmd_list = [self.image_cmd, "measure", ("--output=%s" % output),
                    ("-O %s" % target_fmt)]

        if target_fmt == "luks":
            target_image = self.params.get("image_measure_target", "tgt")
            target_image_secret = self.params.get("image_secret_%s" %
                                                  target_image, "measure")
            target_image_params = self.params.object_params(target_image)
            target_image_params["image_format"] = "luks"
            target_image_params["image_secret"] = target_image_secret
            target_image_object = QemuImg(
                    target_image_params, self.root_dir, target_image)
            cmd_list.append(target_image_object._secret_objects[-1])
            cmd_list.append('-o key-secret=%s' %
                            target_image_object.encryption_config.key_secret.aid)

        if size:
            cmd_list.append(("--size %s" % size))
        else:
            if self.encryption_config.key_secret:
                cmd_list.append(self._secret_objects[-1])
                image_json_str = get_image_json(self.tag,
                                                self.params,
                                                self.root_dir)
                cmd_list.append("'%s'" % image_json_str)
            else:
                cmd_list.extend([("-f %s" % self.image_format),
                                self.image_filename])

        cmd_result = process.run(" ".join(cmd_list), ignore_status=True)
        return cmd_result

    def dd(self, output, bs=None, count=None, skip=None):
        """
        Qemu image dd wrapper, like dd command, clone the image.
        Please use convert to convert one format of image to another.
        :param output: of=output
        :param bs: bs=bs, the block size in bytes
        :param count: count=count, count of blocks copied
        :param skip: skip=skip, count of blocks skipped
        :return: process.CmdResult object containing the result of the
                 command
        """
        cmd_dict = {
            "image_filename": self.image_filename,
            "target_image_filename": output,
            "image_format": self.image_format,
            "target_image_format": self.image_format
        }

        cmd_dict['block_size'] = 'bs=%d' % bs if bs is not None else ''
        cmd_dict['count'] = 'count=%d' % count if count is not None else ''
        cmd_dict['skip'] = 'skip=%d' % skip if skip is not None else ''

        # TODO: use raw copy(-f raw -O raw) and ignore image secret and format
        # for we cannot set secret for the output
        raw_copy = True if self.encryption_config.key_secret else False
        if raw_copy:
            cmd_dict['image_format'] = cmd_dict['target_image_format'] = 'raw'

        # use 'json:{}' instead when accessing storage with auth
        meta = _get_image_meta(self.tag,
                               self.params,
                               self.root_dir) if self._need_auth_info(self.tag) else None
        if meta is not None:
            if raw_copy:
                # drop image secret from meta
                for key in ['encrypt.key-secret', 'key-secret']:
                    if key in meta:     # pylint: disable=E1135
                        meta.pop(key)
            meta['driver'] = cmd_dict.pop("image_format")   # pylint: disable=E1137
            cmd_dict["image_filename"] = "'json:%s'" % json.dumps(meta)

        # access secret objects of the backing images
        secret_objects = self._backing_access_secret_objects

        # access secret object of the image itself
        if self._image_access_secret_object:
            secret_objects.extend(self._image_access_secret_object)

        if secret_objects:
            cmd_dict["secret_object"] = " ".join(secret_objects)

        # tls creds objects of the backing images
        tls_creds_objects = self._backing_access_tls_creds_objects

        # tls creds object of the image itself
        if self._image_access_tls_creds_object:
            tls_creds_objects.append(self._image_access_tls_creds_object)

        if tls_creds_objects:
            cmd_dict["tls_creds_object"] = " ".join(tls_creds_objects)

        dd_cmd = self.image_cmd + " " + \
            self._cmd_formatter.format(self.dd_cmd, **cmd_dict)

        return process.run(dd_cmd, ignore_status=True)

    def copy_data_remote(self, src, dst):
        bs = 1024 * 1024  # 1M, faster copy
        self.dd(dst, bs)


class Iscsidev(storage.Iscsidev):

    """
    Class for handle iscsi devices for VM
    """

    def __init__(self, params, root_dir, tag):
        """
        Init the default value for image object.

        :param params: Dictionary containing the test parameters.
        :param root_dir: Base directory for relative filenames.
        :param tag: Image tag defined in parameter images
        """
        super(Iscsidev, self).__init__(params, root_dir, tag)

    def setup(self):
        """
        Access the iscsi target. And return the local raw device name.
        """
        if self.iscsidevice.logged_in():
            logging.warn("Session already present. Don't need to"
                         " login again")
        else:
            self.iscsidevice.login()

        if utils_misc.wait_for(self.iscsidevice.get_device_name,
                               self.iscsi_init_timeout):
            device_name = self.iscsidevice.get_device_name()
        else:
            raise exceptions.TestError("Can not get iscsi device name in host"
                                       " in %ss" % self.iscsi_init_timeout)

        if self.device_id:
            device_name += self.device_id
        return device_name

    def cleanup(self):
        """
        Logout the iscsi target and clean up the config and image.
        """
        if self.exec_cleanup:
            self.iscsidevice.cleanup()
            if self.emulated_file_remove:
                logging.debug("Removing file %s", self.emulated_image)
                if os.path.exists(self.emulated_image):
                    os.unlink(self.emulated_image)
                else:
                    logging.debug("File %s not found", self.emulated_image)


class LVMdev(storage.LVMdev):

    """
    Class for handle lvm devices for VM
    """

    def __init__(self, params, root_dir, tag):
        """
        Init the default value for image object.

        :param params: Dictionary containing the test parameters.
        :param root_dir: Base directory for relative filenames.
        :param tag: Image tag defined in parameter images
        """
        super(LVMdev, self).__init__(params, root_dir, tag)

    def setup(self):
        """
        Get logical volume path;
        """
        return self.lvmdevice.setup()

    def cleanup(self):
        """
        Cleanup useless volumes;
        """
        return self.lvmdevice.cleanup()



"""
Windows virtio-win utilities
"""

import re
import logging

from . import drive
from . import system
from avocado.core import exceptions


ARCH_MAP_ISO = {"32-bit": "x86", "64-bit": "amd64"}
ARCH_MAP_VFD = {"32-bit": "i386", "64-bit": "amd64"}


def arch_dirname_iso(session):
    """
    Get architecture directory's name - iso media version.

    :param session: Session object.

    :return: Directory's name.
    """
    return ARCH_MAP_ISO.get(system.os_arch(session))


def arch_dirname_vfd(session):
    """
    Get architecture directory's name - vfd media version.

    :param session: Session object.

    :return: Directory's name.
    """
    return ARCH_MAP_VFD.get(system.os_arch(session))


def _product_info(session):
    # Some windows system would present 'r' as the registered
    # trademark character at the end of string "Server"
    match = re.search(r"Windows((?: )Serverr?)? (\S+)(?: (R2))?",
                      system.product_name(session), re.I)
    if not match:
        return ("", "", "")
    server, name, suffix = match.groups()
    server = server if server else ""
    suffix = suffix if suffix else ""
    return server, name, suffix


def product_dirname_iso(session):
    """
    Get product directory's name - iso media version.

    :param session: Session object.

    :return: Directory's name.
    """
    server, name, suffix = _product_info(session)
    if not name:
        return None
    if server:
        if len(name) == 4:
            name = re.sub("0+", "k", name)
    else:
        if name[0].isdigit():
            name = "w" + name
    return name + suffix


def product_dirname_vfd(session):
    """
    Get product directory's name - vfd media version.

    :param session: Session object.

    :return: Directory's name.
    """
    server, name, suffix = _product_info(session)
    if not name:
        return None
    return "Win" + name + suffix


def drive_letter_iso(session):
    """
    Get virtio-win drive letter - iso media version.

    :param session: Session object.

    :return: Drive letter.
    """
    return drive.get_hard_drive_letter(session, "virtio-win%")


def drive_letter_vfd(session):
    """
    Get virtio-win drive letter - vfd media version.

    :param session: Session object.

    :return: Drive letter.
    """
    for letter in drive.get_floppy_drives_letter(session):
        # FIXME: addresses the drive accurately
        return letter
    return None


def _get_netkvmco_path(session):
    """
    Get the proper netkvmco.dll path from iso.

    :param session: a session to send cmd
    :return: the proper netkvmco.dll path.
    """

    viowin_ltr = drive_letter_iso(session)
    if not viowin_ltr:
        err = "Could not find virtio-win drive in guest"
        raise exceptions.TestError(err)
    guest_name = product_dirname_iso(session)
    if not guest_name:
        err = "Could not get product dirname of the vm"
        raise exceptions.TestError(err)
    guest_arch = arch_dirname_iso(session)
    if not guest_arch:
        err = "Could not get architecture dirname of the vm"
        raise exceptions.TestError(err)

    middle_path = "%s\\%s" % (guest_name, guest_arch)
    find_cmd = 'dir /b /s %s\\netkvmco.dll | findstr "\\%s\\\\"'
    find_cmd %= (viowin_ltr,  middle_path)
    netkvmco_path = session.cmd(find_cmd).strip()
    logging.info("Found netkvmco.dll file at %s" % netkvmco_path)
    return netkvmco_path


def prepare_netkvmco(vm):
    """
    Copy the proper netkvmco.dll to driver c:\\, and register it.

    param vm: the target vm
    """
    logging.info("Prepare the netkvmco.dll")
    session = vm.wait_for_login(timeout=360)
    try:
        netkvmco_path = _get_netkvmco_path(session)
        prepare_netkvmco_cmd = "xcopy %s c:\\ /y && "
        prepare_netkvmco_cmd += "rundll32 netkvmco.dll,"
        prepare_netkvmco_cmd += "RegisterNetKVMNetShHelper"
        session.cmd(prepare_netkvmco_cmd % netkvmco_path, timeout=240)
    finally:
        session.close()



"""
redirdev device support class(es)

http://libvirt.org/formatdomain.html#elementsRedir
"""

from virttest.libvirt_xml import accessors
from virttest.libvirt_xml.devices import base, librarian


class Redirdev(base.TypedDeviceBase):

    __slots__ = ('type', 'bus', 'source', 'protocol', 'boot', 'alias', 'address')

    def __init__(self, type_name="redirdev", virsh_instance=base.base.virsh):
        super(Redirdev, self).__init__(device_tag='redirdev',
                                       type_name=type_name,
                                       virsh_instance=virsh_instance)
        accessors.XMLAttribute('type', self, parent_xpath='/',
                               tag_name='redirdev', attribute='type')
        accessors.XMLAttribute('bus', self, parent_xpath='/',
                               tag_name='redirdev', attribute='bus')
        accessors.XMLElementNest('source', self, parent_xpath='/',
                                 tag_name='source', subclass=self.Source,
                                 subclass_dargs={
                                     'virsh_instance': virsh_instance})
        accessors.XMLElementDict('protocol', self, parent_xpath='/',
                                 tag_name='protocol')
        accessors.XMLElementDict('boot', self, parent_xpath='/',
                                 tag_name='boot')
        accessors.XMLElementDict('alias', self, parent_xpath='/',
                                 tag_name='alias')
        accessors.XMLElementNest('address', self, parent_xpath='/',
                                 tag_name='address', subclass=self.Address,
                                 subclass_dargs={'type_name': 'usb',
                                                 'virsh_instance': virsh_instance})
    Address = librarian.get('address')

    def new_source(self, **dargs):
        """
        Return a new Redirdev Source instance and set properties from dargs
        """
        new_one = self.Source(virsh_instance=self.virsh)
        for k, v in dargs.items():
            if v:
                setattr(new_one, k, v)
        return new_one

    class Source(base.base.LibvirtXMLBase):

        __slots__ = ('mode', 'host', 'service', 'tls', 'path')

        def __init__(self, virsh_instance=base.base.virsh):
            accessors.XMLAttribute('mode', self, parent_xpath='/',
                                   tag_name='source', attribute='mode')
            accessors.XMLAttribute('host', self, parent_xpath='/',
                                   tag_name='source', attribute='host')
            accessors.XMLAttribute('service', self, parent_xpath='/',
                                   tag_name='source', attribute='service')
            accessors.XMLAttribute('tls', self, parent_xpath='/',
                                   tag_name='source', attribute='tls')
            accessors.XMLAttribute('path', self, parent_xpath='/',
                                   tag_name='source', attribute='path')
            super(self.__class__, self).__init__(virsh_instance=virsh_instance)
            self.xml = '<source/>'



import os
import logging

import aexpect
from aexpect import remote

from avocado.utils import process
from avocado.utils import path

from virttest import remote as remote_old


def get_public_key():
    """
    Return a valid string ssh public key for the user executing autoserv or
    autotest. If there's no DSA or RSA public key, create a RSA keypair with
    ssh-keygen and return it.

    :returns: a ssh public key
    :rtype: str
    """

    ssh_conf_path = os.path.expanduser('~/.ssh')

    dsa_public_key_path = os.path.join(ssh_conf_path, 'id_dsa.pub')
    dsa_private_key_path = os.path.join(ssh_conf_path, 'id_dsa')

    rsa_public_key_path = os.path.join(ssh_conf_path, 'id_rsa.pub')
    rsa_private_key_path = os.path.join(ssh_conf_path, 'id_rsa')

    has_dsa_keypair = (os.path.isfile(dsa_public_key_path) and
                       os.path.isfile(dsa_private_key_path))
    has_rsa_keypair = (os.path.isfile(rsa_public_key_path) and
                       os.path.isfile(rsa_private_key_path))

    if has_rsa_keypair:
        logging.info('RSA keypair found, using it')
        public_key_path = rsa_public_key_path

    elif has_dsa_keypair:
        logging.info('DSA keypair found, using it')
        public_key_path = dsa_public_key_path

    else:
        logging.info('Neither RSA nor DSA keypair found, creating RSA ssh key pair')
        process.system('ssh-keygen -t rsa -q -N "" -f %s' %
                       rsa_private_key_path)
        public_key_path = rsa_public_key_path

    public_key = open(public_key_path, 'r')
    public_key_str = public_key.read()
    public_key.close()

    return public_key_str


def get_remote_public_key(session, public_key="rsa"):
    """
    Return a valid string ssh public key for the user executing autoserv or
    autotest. If there's no DSA or RSA public key, create a RSA keypair with
    ssh-keygen and return it.

    :param session: A ShellSession for remote host
    :returns: a ssh public key
    :rtype: str
    """
    session.cmd_output("mkdir -p ~/.ssh")
    session.cmd_output("chmod 700 ~/.ssh")

    ssh_conf_path = "~/.ssh"
    dsa_public_key_path = os.path.join(ssh_conf_path, 'id_dsa.pub')
    dsa_private_key_path = os.path.join(ssh_conf_path, 'id_dsa')

    rsa_public_key_path = os.path.join(ssh_conf_path, 'id_rsa.pub')
    rsa_private_key_path = os.path.join(ssh_conf_path, 'id_rsa')

    dsa_public_s = session.cmd_status("ls %s" % dsa_public_key_path)
    dsa_private_s = session.cmd_status("ls %s" % dsa_private_key_path)
    rsa_public_s = session.cmd_status("ls %s" % rsa_public_key_path)
    rsa_private_s = session.cmd_status("ls %s" % rsa_private_key_path)

    has_dsa_keypair = dsa_public_s == 0 and dsa_private_s == 0
    has_rsa_keypair = rsa_public_s == 0 and rsa_private_s == 0

    if has_dsa_keypair and public_key == "dsa":
        logging.info('DSA keypair found on %s, using it', session)
        public_key_path = dsa_public_key_path

    elif has_rsa_keypair and public_key == "rsa":
        logging.info('RSA keypair found on %s, using it', session)
        public_key_path = rsa_public_key_path

    else:
        logging.info('Neither RSA nor DSA keypair found, '
                     'creating %s ssh key pair' % public_key)
        key_path = rsa_private_key_path
        public_key_path = rsa_public_key_path
        if public_key == "dsa":
            key_path = dsa_private_key_path
            public_key_path = dsa_public_key_path
        session.cmd('ssh-keygen -t %s -q -N "" -f %s' %
                    (public_key, key_path))

    return session.cmd_output("cat %s" % public_key_path)


def setup_ssh_key(hostname, user, password, port=22):
    """
    Setup up remote login in another server by using public key

    :param hostname: the server to login
    :type hostname: str
    :param user: user to login
    :type user: str
    :param password: password
    :type password: str
    :param port: port number
    :type port: int
    """
    logging.debug('Performing SSH key setup on %s:%d as %s.' %
                  (hostname, port, user))

    try:
        session = remote.remote_login(client='ssh', host=hostname,
                                      username=user, port=port,
                                      password=password, prompt=r'[$#%]')
        public_key = get_public_key()

        session.cmd('mkdir -p ~/.ssh')
        session.cmd('chmod 700 ~/.ssh')
        session.cmd("echo '%s' >> ~/.ssh/authorized_keys; " %
                    public_key)
        session.cmd('chmod 600 ~/.ssh/authorized_keys')
        logging.debug('SSH key setup complete.')

    except Exception:
        logging.debug('SSH key setup has failed.')

    finally:
        try:
            session.close()
        except Exception:
            pass


def setup_remote_ssh_key(hostname1, user1, password1,
                         hostname2=None, user2=None, password2=None,
                         port=22, config_options=None, public_key="rsa"):
    """
    Setup up remote to remote login in another server by using public key
    If hostname2 is not supplied, setup to local.

    :param hostname1: the server wants to login other host
    :param hostname2: the server to be logged in
    :type hostname: str
    :param user: user to login
    :type user: str
    :param password: password
    :type password: str
    :param port: port number
    :type port: int
    :param config_options: list of options eg: ["StrictHostKeyChecking=no"]
    :type config_options: list of str
    """
    logging.debug('Performing SSH key setup on %s:%d as %s.' %
                  (hostname1, port, user1))

    try:
        session1 = remote.remote_login(client='ssh', host=hostname1, port=port,
                                       username=user1, password=password1,
                                       prompt=r'[$#%]')
        public_key = get_remote_public_key(session1, public_key=public_key)

        if hostname2 is None:
            # Simply create a session to local
            session2 = aexpect.ShellSession("sh", linesep='\n', prompt='#')
            # set config in local machine
            if config_options:
                for each_option in config_options:
                    session2.cmd_output("echo '%s' >> ~/.ssh/config" %
                                        each_option)
        else:
            session2 = remote.remote_login(client='ssh', host=hostname2,
                                           port=port, username=user2,
                                           password=password2,
                                           prompt=r'[$#%]')
            # set config in remote machine
            if config_options:
                for each_option in config_options:
                    session1.cmd_output("echo '%s' >> ~/.ssh/config" %
                                        each_option)
        session2.cmd_output('mkdir -p ~/.ssh')
        session2.cmd_output('chmod 700 ~/.ssh')
        session2.cmd_output("echo '%s' >> ~/.ssh/authorized_keys; " %
                            public_key)
        session2.cmd_output('chmod 600 ~/.ssh/authorized_keys')
        logging.debug('SSH key setup on %s complete.', session2)
    except Exception as err:
        logging.debug('SSH key setup has failed: %s', err)
        try:
            session1.close()
            session2.close()
        except Exception:
            pass


def setup_remote_known_hosts_file(client_ip, server_ip,
                                  server_user, server_pwd):
    """
    Set the ssh host key of local host to remote host

    :param client_ip: local host ip whose host key is sent to remote host
    :type client_ip: str
    :param server_ip: remote host ip address where host key is stored to
    :type server_ip: str
    :param server_user: user to log on remote host
    :type server_user: str
    :param server_pwd: password for the user for log on remote host
    :type server_pwd: str

    :return: a RemoteFile object for the file known_hosts on remote host
    :rtype: remote_old.RemoteFile
    :return: None if required command is not found
    """
    logging.debug('Performing known_hosts file setup on %s from %s.' %
                  (server_ip, client_ip))
    abs_path = ""
    try:
        abs_path = path.find_command("ssh-keyscan")
    except path.CmdNotFoundError as err:
        logging.debug("Failed to find the command: %s", err)
        return None

    cmd = "%s %s" % (abs_path, client_ip)
    host_key = process.run(cmd, verbose=False).stdout_text
    remote_known_hosts_file = remote_old.RemoteFile(
        address=server_ip,
        client='scp',
        username=server_user,
        password=server_pwd,
        port='22',
        remote_path='~/.ssh/known_hosts')
    pattern2repl = {r".*%s[, ].*" % client_ip: host_key}
    remote_known_hosts_file.sub_else_add(pattern2repl)
    return remote_known_hosts_file
